/***********************************************************************
* mexAsum.c : C mex file to compute sum_{k=1}^m Ak*yk.
*             
*   Ay = mexAsum(blk,A,y,iscmp);  
* 
*   input:  blk = a px2 cell array describing the block structure
*                 of A
*           A = either an pxm  CELL ARRAY where each cell 
*               is a real symmetric matrix, or a 
*               column vector. 
*           y = an mx1 column vector. 
*           iscmp = a binary flag indicate whether Ay is complex
*
*  output:  Ay = a px1 cell array. 
*
* SDPT3: version 2.1 
* Copyright (c) 1997 by
* K.C. Toh, M.J. Todd, R.H. Tutuncu
* Last Modified: 7 Jul 99   
************************************************************************/

#include <math.h>
#include <mex.h>

/**********************************************************
* numblk = 1: matrix + scalar*matrix
***********************************************************/
void  Asum(int n, double *B, double *BI,
                  double alpha, 
                  double *A, double *AI, int *irA, int *jcA, int isspA,
                  int iscmpA, int iscmpB) 
{     
   int  i, j, k, r, jn, idx, kstart, kend;
   double tmp; 

   if (iscmpA & iscmpB) {
      if (!isspA){
         for (j=0; j<n; j++){
            jn = j*n; 
            for (i=0; i<=j; i++){
               idx = i + jn; 
               B[idx]  += alpha*A[idx];
               BI[idx] += alpha*AI[idx];
            }
         } }
      else {
         for (j=0; j<n; j++){
            kend = jcA[j+1]; 
            k = jcA[j];
            r = irA[k];
            while (k < kend) {
               B[r+j*n]  += alpha*A[k]; 
               BI[r+j*n] += alpha*AI[k]; 
               k++;
               r = irA[k]; }
         }
      }
   }
   else {
      if (!isspA){
         for (j=0; j<n; j++){
            jn = j*n; 
            for (i=0; i<=j; i++){
               idx = i + jn; 
               B[idx] += alpha*A[idx];
            }
         } }
      else {
         for (j=0; j<n; j++){
            kend = jcA[j+1]; 
            k = jcA[j];
            r = irA[k];
            while (k < kend) {
               B[r+j*n] += alpha*A[k]; 
               k++;
               r = irA[k]; }
         }
      }
   }            
   return;
}

/**********************************************************
* numblk > 1:  matrix + scalar*matrix
***********************************************************/
void  Asum2(int n, double *B, double *BI, int *irB, int *jcB, int iscmpB,
            double alpha, 
            double *A, double *AI, int *irA, int *jcA, int isspA, int iscmpA,
            double *blksize, int *cumsum, int *blknnz, int numblk, int findirB) 

{ int l, j1, k1, r, idx, cols, colstart, colend,  kstart, kend;  


   if (iscmpA && iscmpB) {
      for (l=0; l<numblk; l++) {
	 /********************************
	 * sub-blocks of a sparse block
	 *******************************/
         cols   = (int)blksize[l]; 
         colstart = cumsum[l]; 
         colend   = cumsum[l+1]; 
	 for (j1=colstart; j1<colend; j1++) {
            kstart = jcA[j1]; 
            kend = jcA[j1+1];
            idx = (j1-colstart)*cols + blknnz[l]; 
   	    for (k1=kstart; k1<kend; k1++) {
               r = irA[k1]-colstart;
               B[idx+r]  += alpha*A[k1];
               BI[idx+r] += alpha*AI[k1]; }
	 }
      }
   }
   else {
      for (l=0; l<numblk; l++) {
         cols   = (int)blksize[l]; 
         colstart = cumsum[l];
         colend   = cumsum[l+1]; 
	 for (j1=colstart; j1<colend; j1++) {
            kstart = jcA[j1]; 
            kend = jcA[j1+1];
            idx = (j1-colstart)*cols + blknnz[l]; 
   	    for (k1=kstart; k1<kend; k1++) {
               r = irA[k1]-colstart;
               B[idx+r] += alpha*A[k1]; }
	 }
      }
   }
   if (findirB) {
      idx = 0; 
      for (l=0; l<numblk; l++) {
         cols = (int)blksize[l]; 
         colstart = cumsum[l];
         colend   = cumsum[l+1];
         for (j1=colstart; j1<colend; j1++) {
            jcB[j1] = (j1-colstart)*cols + blknnz[l];
            for (k1=colstart; k1<colend; k1++){
               irB[idx] = k1;      
               idx++; }
	 }
      }
      jcB[n] = blknnz[numblk]; 
   }
return;
}
/**********************************************************
* column + scalar*column
***********************************************************/
void  Asumdiag(int n, double *B,  
               double alpha, 
               double *A, int *irA, int *jcA, int isspA) 
{     
   int  i, k, r, kstart, kend;
   double tmp; 

   if (!isspA){
      for (i=0; i<n; i++){
          B[i] += alpha*A[i];
      }
   }
   else {
      kstart = jcA[0];
      kend = jcA[1]; 
      for (k=kstart; k<kend; ++k){
         r = irA[k];
         B[r] += alpha*A[k];
      }
   }        
   return;
}
/**********************************************************
* form P using the upper triangular part
**********************************************************/
void symmetrize(int n, double *P)
{
   int j, k, indj, indk; 

   for (j=0; j<n; ++j){
      indj = j*n; 
      for (k=0; k<j ; ++k){ 
         indk = k*n;
         P[j+indk] = P[k+indj];
      }
   }
   return;
}
/**********************************************************/
void antisymmetrize(int n, double *P)
{
   int j, k, indj, indk; 

   for (j=0; j<n; ++j){
      indj = j*n; 
      for (k=0; k<j ; ++k){ 
         indk = k*n;
         P[j+indk] = -P[k+indj];
      }
   }
   return;
}
/**********************************************************
* 
***********************************************************/
void mexFunction(
      int nlhs,   mxArray  *plhs[], 
      int nrhs,   const mxArray  *prhs[] )
{    
     mxArray  *A_cell_pr, *blk_cell_pr, *lhs;
     double   *A, *B, *AI, *BI, *y;
     double   *blksize, *cmpdata;
     int      *cumsum, *blknnz;
     int      *irA, *jcA, *irB, *jcB;
     double   alpha;
     int      iscellA, mA, nA, my, ny, isspA, iscmpA, iscmpB, findirB;

     int   subs[2];
     int   nsubs=2; 
     int   m1, n1, m2, n2, i, j, k, l, r, index, i1, j1, k1, jn, idx;
     int   numblk, cols, colstart, colend, NZmax, kstart, kend;

/* CHECK FOR PROPER NUMBER OF ARGUMENTS */

   if (nrhs < 3){
      mexErrMsgTxt("mexAsum: requires 3 or more input arguments."); }
   else if (nlhs>1){ 
      mexErrMsgTxt("mexAsum: requires 1 output argument."); }

/* CHECK THE DIMENSIONS */

   mA = mxGetM(prhs[1]);
   nA = mxGetN(prhs[1]); 
   my = mxGetM(prhs[2]); 
   ny = mxGetN(prhs[2]);    
   iscellA = mxIsCell(prhs[1]); 

   if (ny > 1){
      mexErrMsgTxt("mexAsum: 2ND input must be a column vector"); }
   if (nA != my){
      mexErrMsgTxt("mexAsum: dimensions of 1ST and 2ND input not compatible");}
   if (!iscellA){
       mexErrMsgTxt("mexAsum: 1ST input must be a cell array"); }

/* Create a matrix for the return argument and assign pointers */

   y = mxGetPr(prhs[2]); 
   plhs[0] = mxCreateCellMatrix(mA,1);
   
/* Do the computations in a subroutine */

       if (nrhs==3){
          iscmpB = 0; }
       else {
          cmpdata = mxGetPr(prhs[3]); 
       }
      /**************************************************
       * summing each row of the cell array
       *************************************************/    
       for (i=0; i<mA; i++){
           subs[0] = i; 
           subs[1] = 0;
           index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
           A_cell_pr = mxGetCell(prhs[1],index); 
   	   m1 = mxGetM(A_cell_pr); 
           n1 = mxGetN(A_cell_pr);
           subs[0] = i; 
           subs[1] = 1;
           index = mxCalcSingleSubscript(prhs[0],nsubs,subs); 
           blk_cell_pr = mxGetCell(prhs[0],index);
           numblk  = mxGetN(blk_cell_pr);            
           blksize = mxGetPr(blk_cell_pr);
           if (nrhs > 3) { iscmpB = (int) cmpdata[i]; }
           if (numblk == 1){
              if (!iscmpB){ 
                 lhs = mxCreateDoubleMatrix(m1,n1,mxREAL); }
              else {
                 lhs = mxCreateDoubleMatrix(m1,n1,mxCOMPLEX); 
                 BI = mxGetPi(lhs); 
              }
              B = mxGetPr(lhs); }
           else { 
              cumsum = mxCalloc(numblk+1,sizeof(int)); 
              blknnz = mxCalloc(numblk+1,sizeof(int)); 
              cumsum[0] = 0; blknnz[0] = 0;
 	      for (k=0; k<numblk; k++) {
                  cols = (int)blksize[k]; 
		  blknnz[k+1] = blknnz[k]+ cols*cols;
                  cumsum[k+1] = cumsum[k] + cols; }
              m1 = cumsum[numblk];
              NZmax = blknnz[numblk]; 
              if (!iscmpB){
		 lhs = mxCreateSparse(m1,m1,NZmax,mxREAL); }
              else {
                 lhs = mxCreateSparse(m1,m1,NZmax,mxCOMPLEX); 
                 BI = mxGetPi(lhs); }
              B = mxGetPr(lhs); 
              irB = mxGetIr(lhs); 
              jcB = mxGetJc(lhs); 
           }
	   /******************************************
	    * summing yj*Aj over j
	    *****************************************/
           for (j=0; j<nA; j++){
               alpha = y[j]; 
               subs[0] = i; 
               subs[1] = j;
               index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
               A_cell_pr = mxGetCell(prhs[1],index); 
               A = mxGetPr(A_cell_pr); 
               isspA  = mxIsSparse(A_cell_pr); 
               iscmpA = mxIsComplex(A_cell_pr);  
    	       m2 = mxGetM(A_cell_pr);
    	       n2 = mxGetN(A_cell_pr);
               if (isspA) { 
                  irA = mxGetIr(A_cell_pr); 
                  jcA = mxGetJc(A_cell_pr);
               }
               if (iscmpA) { AI = mxGetPi(A_cell_pr); }
               if (numblk == 1 && n2 > 1) {
                  if (alpha != 0) { 
  	  	     Asum(m2,B,BI,alpha,A,AI,irA,jcA,isspA,iscmpA,iscmpB); }
                  if (j == nA-1) {
 		     symmetrize(m2,B);  
                     if (iscmpB) { antisymmetrize(m2,BI); }
                  }
               }
	       else if (numblk == 1 && n2 == 1) {
		  if (alpha != 0) {
		     Asumdiag(m2,B,alpha,A,irA,jcA,isspA); }
               }
               else {
 		  if (!isspA) { 
                     mexErrMsgTxt("mexAsum: numblk > 1, data must be in sparse format."); }
                  findirB = (j == nA-1);                  
		  if (alpha != 0 || findirB) {
                     Asum2(m2,B,BI,irB,jcB,iscmpB,alpha,A,AI,irA,jcA,isspA,iscmpA, \
                           blksize,cumsum,blknnz,numblk,findirB);                   
		  }
	       }
	   }
	  /**************************************
	   * put result in the ith cell
           *************************************/
          subs[0] = i; 
          subs[1] = 0; 
          index = mxCalcSingleSubscript(plhs[0],nsubs,subs); 
          mxSetCell(plhs[0],index,lhs); 
       }
    return; 
}
/**********************************************************/

