/******************************************************************
* mexProd2.c : C-MEX file to compute the product of two matrices. 
*       
*         P = mexProd2(blk,A,B,type)
*
* input:  blk = cell array describing the block structure of A and B. 
*         A = nxn matrix.
*         B = nxn matrix.
*      type = 0     general matrix product
*             1     if P is hermitian
*             21    if                        B is upper triangular 
*             211   if        B is upper triangular, P is hermitian
*             22    if                        B is lower triangular
*             221   if        B is lower triangular, P is hermitian
*             31    if A is upper triangular
*             311   if A is upper triangular,        P is hermitian
*             32    if A is lower triangular
*             321   if A is lower triangular,        P is hermitian
*             41    if A is upper triangular, B is upper triangular
*             42    if A is upper triangular, B is lower triangualr
*             421   if A is upper triangular, B is lower triangualr,
*                                                    P is hermitian
*
*  IMPORTANT: when A and B are sparse matrices, and P is hermitian,
*             only the upper triangular part of P is returned. 
*                      
* SDPT3: version 2.1 
* Copyright (c) 1997 by
* K.C. Toh, M.J. Todd, R.H. Tutuncu
* Last Modified: 7 Jul 99
******************************************************************/

#include <math.h>
#include <mex.h>

static int min1, min2;
#define IMIN(a,b) (min1=(a),min2=(b),(min1)<(min2)?(min1):(min2))

/**********************************************************
* form P using the upper triangular part
**********************************************************/
void symmetrize(double *P, int n)
{
   int j, k, jn, kn; 

   for (j=0; j<n; ++j){
       jn = j*n; 
       for (k=0; k<j ; ++k){ 
           kn = k*n;
           P[j+kn] = P[k+jn]; }
   }
   return;
}
void antisymmetrize(double *P, int n)
{
   int j, k, jn, kn; 

   for (j=0; j<n; ++j){
       jn = j*n; 
       for (k=0; k<j ; ++k){ 
           kn = k*n;
           P[j+kn] = -P[k+jn]; }
       P[j+jn] = 0; 
   }
   return;
}
/********************************************************
* determine istart, iend
********************************************************/
void product_idx(int index[2], int j, int k, int n, int type)

{   int istart, iend;
  
    if (type == 1 || type == 211 || type == 221)
       { index[0] = 0; index[1] = j; }
    else if (type == 0 || type == 21 || type == 22)
       { index[0] = 0;  index[1] = n-1; }
    else if (type == 31 || type == 41 || type == 42)
       { index[0] = 0; index[1] = k; }
    else if (type == 311 || type == 421) 
       { index[0] = 0; index[1] = IMIN(k,j); }
    else if (type == 32)
       { index[0] = k; index[1] = n-1; }
    else if (type == 321)
       { index[0] = k; index[1] = j; }
    return;
}
/**********************************************************
* A dense, B dense
**********************************************************/
static void product(double *A, double *AI, double *B, double *BI,
                    double *P, double *PPI, 
                    int n, int type, int iscmpA, int iscmpB)

{  int	idx, i, j, k, jn, kn, kstart, kend, istart, iend;
   int  index[2]; 
   double tmp, tmPPI;

   if (!iscmpA) {
      for (j=0; j<n; ++j){
          if (type==0 || type==1 || type==31 || type==311 || type==32)
             { kstart = 0;  kend = n-1; }
          else if (type == 22 || type == 221 || type == 42 || type == 421)
	     { kstart = j;  kend = n-1; }
          else if (type == 21 || type == 211 || type == 321 || type == 41)
	     { kstart = 0;  kend = j; } 
          jn = j*n; 
	  for (k=kstart; k<=kend; ++k){
              product_idx(index, j, k, n, type); 
              istart = index[0]; iend = index[1];
              kn = k*n;
              tmp  = B[k+jn]; 
              if (!iscmpB) { tmPPI = 0; } else { tmPPI = BI[k+jn]; }
              if (tmp != 0 && tmPPI == 0) {
                 for (i=istart; i<=iend; ++i){
                     idx = i+kn; 
		     P[i+jn]  += tmp*A[idx]; }
              }
              else if (tmp != 0 && tmPPI != 0) {
                 for (i=istart; i<=iend; ++i){
                     idx = i+kn; 
		     P[i+jn]  += tmp*A[idx]; 
		     PPI[i+jn] += tmPPI*A[idx]; }              
              }
              else if (tmp == 0 && tmPPI != 0) {
                 for (i=istart; i<=iend; ++i){
                     idx = i+kn; 
		     PPI[i+jn] += tmPPI*A[idx]; }              
              }
          }
      }
   }
   else if (iscmpA) {
      for (j=0; j<n; ++j){
          if (type==0 || type==1 || type==31 || type==311 || type==32)
             { kstart = 0;  kend = n-1; }
          else if (type == 22 || type == 221 || type == 42 || type == 421)
	     { kstart = j;  kend = n-1; }
          else if (type == 21 || type == 211 || type == 321 || type == 41)
	     { kstart = 0;  kend = j; } 
          jn = j*n; 
	  for (k=kstart; k<=kend; ++k){
              product_idx(index, j, k, n, type); 
              istart = index[0]; iend = index[1];
              kn = k*n;
              tmp  = B[k+jn]; 
              if (!iscmpB) { tmPPI = 0; } else { tmPPI = BI[k+jn]; }
              if (tmp != 0 && tmPPI != 0) {
                 for (i=istart; i<=iend; ++i){
                     idx = i+kn; 
		     P[i+jn]  += tmp*A[idx]  - tmPPI*AI[idx]; 
		     PPI[i+jn] += tmp*AI[idx] + tmPPI*A[idx]; }              
              }
              else if (tmp != 0 && tmPPI == 0) {
                 for (i=istart; i<=iend; ++i){
                     idx = i+kn; 
		     P[i+jn]  += tmp*A[idx]; 
		     PPI[i+jn] += tmp*AI[idx]; }
              }
              else if (tmp == 0 && tmPPI != 0) {
                 for (i=istart; i<=iend; ++i){
                     idx = i+kn; 
		     P[i+jn]  -= tmPPI*AI[idx]; 
		     PPI[i+jn] += tmPPI*A[idx]; }              
              }
          }
      }
   }
   if (type == 1   || type == 211 || type == 221 || type == 311 || \
       type == 321 || type == 421) {
      if (!iscmpA & !iscmpB)
         { symmetrize(P,n); } 
      else
         { symmetrize(P,n); antisymmetrize(PPI,n); } 
   }
   return; 
}
/**********************************************************
* A dense, B sparse
**********************************************************/
static void product2(double *A, double *AI, 
                     double *B, double *BI, int *irB, int *jcB, 
                     double *P, double *PPI, 
                     int n, int type, int iscmpA, int iscmpB)

{  int	  idx, i, j, k, r, kstart, kend, istart, iend, jn, rn;
   int    index[2];
   double tmp, tmPPI;
    
   if (!iscmpA) {
      for (j=0; j<n; ++j){
          jn = j*n;
          kstart = jcB[j]; 
          kend   = jcB[j+1]-1; 
	  for (k=kstart; k<=kend; ++k){
              r = irB[k]; 
              rn = r*n;
              tmp  = B[k];
              if (!iscmpB) { tmPPI = 0; }  else { tmPPI = BI[k]; }
              product_idx(index, j, r, n, type); 
              istart = index[0]; iend = index[1];
              if (tmp != 0 && tmPPI == 0) {
                 for (i=istart; i<=iend; ++i){
                    idx = i+rn; 
	  	    P[i+jn]  += tmp*A[idx]; }
              }
              else if (tmp != 0 && tmPPI !=0) {
                 for (i=istart; i<=iend; ++i){
                    idx = i+rn; 
	  	    P[i+jn]  += tmp*A[idx]; 
	  	    PPI[i+jn] += tmPPI*A[idx]; }
              }
              else if (tmp == 0 && tmPPI != 0) {
                 for (i=istart; i<=iend; ++i){
                    idx = i+rn; 
	  	    PPI[i+jn] += tmPPI*A[idx]; }
              }
          }
      }
   }
   else if (iscmpA) {
      for (j=0; j<n; ++j){
          jn = j*n;
          kstart = jcB[j]; 
          kend   = jcB[j+1]-1; 
	  for (k=kstart; k<=kend; ++k){
              r = irB[k]; 
              rn = r*n;
              tmp  = B[k];
              if (!iscmpB) { tmPPI = 0; } else { tmPPI = BI[k];}
              product_idx(index, j, r, n, type); 
              istart = index[0]; iend = index[1];
              if (tmp != 0 && tmPPI !=0) {
                 for (i=istart; i<=iend; ++i){
                    idx = i+rn; 
	  	    P[i+jn]  += tmp*A[idx]  - tmPPI*AI[idx]; 
	  	    PPI[i+jn] += tmp*AI[idx] + tmPPI*A[idx]; }
              }
              else if (tmp != 0 && tmPPI == 0) {
                 for (i=istart; i<=iend; ++i){
                    idx = i+rn; 
	  	    P[i+jn]  += tmp*A[idx]; 
	  	    PPI[i+jn] += tmp*AI[idx]; }
              }
              else if (tmp == 0 && tmPPI != 0) {
                 for (i=istart; i<=iend; ++i){
                    idx = i+rn; 
	  	    P[i+jn]  -= tmPPI*AI[idx]; 
	  	    PPI[i+jn] += tmPPI*A[idx]; }
              }
          }
      }
   }
   if (type == 1   || type == 211 || type == 221 || type == 311 || \
       type == 321 || type == 421) { 
     if (!iscmpA & !iscmpB) { symmetrize(P,n); }
     else { symmetrize(P,n); antisymmetrize(PPI,n); }
   }
   return;
}
/**********************************************************
* A sparse, B dense
**********************************************************/
void product3_idx(int index[3], int j, int n, int type)

{ int kstart, kend, sym; 

    if (type == 1 || type == 311 || type == 321)  
       { kstart = 0; kend = n-1; sym = 1; }
    else if (type == 211)
       { kstart = 0; kend = j; sym = 1; }
    else if (type == 221 || type == 421) 
       { kstart = j; kend = n-1; sym = 1; }
    else if (type == 0 || type == 31 || type == 32) 
       { kstart = 0; kend = n-1; sym = 0; }
    else if (type == 21 || type == 41)         
       { kstart = 0; kend = j;   sym = 0; }
    else if (type == 22 || type == 42)         
       { kstart = j; kend = n-1; sym = 0; }
    index[0] = kstart; 
    index[1] = kend; 
    index[2] = sym;
    return;
}
/**********************************************************/
static void product3(double *A, double *AI, int *irA, int *jcA, 
                     double *B, double *BI,   
                     double *P, double *PPI, 
                     int n, int type, int iscmpA, int iscmpB)

{  int	  idx, i, j, k, ri, kstart, kend, istart, iend, jn, kn, sym;
   int    index[3];
   double tmp, tmPPI;
 
   if (!iscmpA) {
      for (j=0; j<n; ++j){
          product3_idx(index, j, n, type); 
          kstart = index[0];  kend = index[1];  sym = index[2]; 
          jn = j*n;
	  for (k=kstart; k<=kend; ++k){
              tmp  = B[k+jn];
              if (!iscmpB) { tmPPI = 0; }  else { tmPPI = BI[k+jn]; }
              istart = jcA[k]; 
              iend = jcA[k+1]-1; 
              if (tmp != 0 && tmPPI == 0){
  		 for (i=istart; i<=iend; ++i) {
                    ri = irA[i]; 
                    if (ri > j & sym) { break; }
		    P[ri+jn]  += tmp*A[i]; }
              }
              else if (tmp != 0 && tmPPI != 0){
  		 for (i=istart; i<=iend; ++i) {
                    ri = irA[i]; 
                    if (ri > j && sym) { break; }
		    P[ri+jn]  += tmp*A[i]; 
                    PPI[ri+jn] += tmPPI*A[i]; }
	      }
              else if (tmp == 0 && tmPPI != 0) {
  		 for (i=istart; i<=iend; ++i) {
                    ri = irA[i]; 
                    if (ri > j & sym) { break; }
                    PPI[ri+jn] += tmPPI*A[i]; }
              }
	  }
      }
   }
   else if (iscmpA) {
      for (j=0; j<n; ++j){
          product3_idx(index, j, n, type); 
          kstart = index[0];  kend = index[1];  sym = index[2]; 
          jn = j*n;
          for (k=kstart; k<=kend; ++k){
              tmp  = B[k+jn];
              if (!iscmpB) { tmPPI = 0; } else { tmPPI = BI[k+jn]; }
              istart = jcA[k]; 
              iend = jcA[k+1]-1; 
              if (tmp != 0 && tmPPI != 0){
  		 for (i=istart; i<=iend; ++i) {
                    ri = irA[i]; 
                    if (ri > j & sym) { break; }
		    P[ri+jn]  += tmp*A[i]  - tmPPI*AI[i]; 
                    PPI[ri+jn] += tmp*AI[i] + tmPPI*A[i]; }
	      }
              else if (tmp != 0 && tmPPI == 0){
  		 for (i=istart; i<=iend; ++i) {
                    ri = irA[i]; 
                    if (ri > j & sym) { break; }
	  	    P[ri+jn]  += tmp*A[i]; 
	  	    PPI[ri+jn] += tmp*AI[i]; }             
              }
              else if (tmp == 0 && tmPPI != 0) {
  		 for (i=istart; i<=iend; ++i) {
                    ri = irA[i]; 
                    if (ri > j & sym) { break; }
	  	    P[ri+jn]  -= tmPPI*AI[i]; 
	  	    PPI[ri+jn] += tmPPI*A[i]; }             
              }
	  }
      }
   }
   if (type == 1   || type == 211 || type == 221 || type == 311 || \
       type == 321 || type == 421) {
      if (!iscmpA & !iscmpB) { symmetrize(P,n); }
      else { symmetrize(P,n); antisymmetrize(PPI,n); }
   }
   return;
}
/**********************************************************
* A sparse, B sparse
**********************************************************/
static void product4(double *A, double *AI, int *irA, int *jcA, 
                     double *B, double *BI, int *irB, int *jcB, 
                     double *P, double *PPI, int *irP, int *jcP,
                     double *Ptmp, double *PPItmp, 
                     int *cumsum, int numblk, 
                     int type, int iscmpA, int iscmpB)

{  int     i, j, k, l, r, t, istart, iend, kstart, kend, colstart, colend;
   int     idx, sym;  
   double  tmp, tmPPI; 

   if (type == 1 || type == 211 || type == 221 || type == 311 || \
       type == 321 || type == 421)
      { sym = 1; } 
   else  
      { sym = 0; }
   idx = 0;  jcP[0]=0; 
   for (l=0; l<numblk; ++l){
      colstart = cumsum[l]; 
      colend = cumsum[l+1]; 
      for (j=colstart; j<colend; ++j){
         kstart = jcB[j]; 
         kend   = jcB[j+1]-1;
         /**** forming jth column of P ****/
         if (!iscmpA) {
            for (k=kstart; k<=kend; ++k) {
               r = irB[k];
               tmp  = B[k];
               if (!iscmpB) { tmPPI = 0; }  else { tmPPI = BI[k]; }
               istart = jcA[r]; 
               iend   = jcA[r+1]-1;
               if (tmp !=0 && tmPPI == 0) {
                  for (i=istart; i<=iend; ++i) {
                     t = irA[i];
                     if (t > j & sym) { break; }
	   	     Ptmp[t]  += tmp*A[i]; }
	       }
               else if (tmp != 0 && tmPPI != 0){
                  for (i=istart; i<=iend; ++i) {
                     t = irA[i];
                     if (t > j & sym) { break; }
	   	     Ptmp[t]  += tmp*A[i]; 
		     PPItmp[t] += tmPPI*A[i];  }
	       }
               else if (tmp ==0 && tmPPI != 0) {
                  for (i=istart; i<=iend; ++i) {
                     t = irA[i]; 
                     if (t > j & sym) { break; }
		     PPItmp[t] += tmPPI*A[i];  }
	       }
	    }          
	 }
         else if (iscmpA) {
            for (k=kstart; k<=kend; ++k) {
               r = irB[k];
               tmp  = B[k];
               if (!iscmpB) { tmPPI = 0; }  else { tmPPI = BI[k]; }
               istart = jcA[r]; 
               iend   = jcA[r+1]-1;
               if (tmp != 0 && tmPPI != 0){
                  for (i=istart; i<=iend; ++i) {
                     t = irA[i];
                     if (t > j & sym) { break; }
	   	     Ptmp[t]  += tmp*A[i]  - tmPPI*AI[i]; 
		     PPItmp[t] += tmp*AI[i] + tmPPI*A[i];  }
	       }
               else if (tmp !=0 && tmPPI == 0) {
                  for (i=istart; i<=iend; ++i) {
                     t = irA[i];
                     if (t > j & sym) { break; }
	   	     Ptmp[t]  += tmp*A[i]; 
		     PPItmp[t] += tmp*AI[i];  }
	       }
               else if (tmp ==0 && tmPPI != 0) {
                  for (i=istart; i<=iend; ++i) {
                     t = irA[i];
                     if (t > j & sym) { break; }
	   	     Ptmp[t]  -= tmPPI*AI[i]; 
		     PPItmp[t] += tmPPI*A[i];  }
	       }
	    }
	 }
         if (kstart <= kend) {
   	    if (!iscmpA & !iscmpB){
               for (k=colstart; k<colend; ++k) {
                  tmp = Ptmp[k]; 
	          if (tmp != 0) {
                     P[idx] = tmp;
                     irP[idx] = k;
                     Ptmp[k] = 0; 
                     idx++; }
	       }
	    }
            else {
               for (k=colstart; k<colend; ++k) {
                  tmp  = Ptmp[k]; 
                  tmPPI = PPItmp[k]; 
	          if (tmp != 0 || tmPPI != 0) {
                     P[idx]   = tmp;
                     PPI[idx]  = tmPPI;
                     irP[idx] = k;
                     Ptmp[k]  = 0;  
                     PPItmp[k] = 0;  
                     idx++; }
	       }
	    }            
	 }
         jcP[j+1] = idx;    
      }
   }
   jcP[colend] = idx; 
   return; 
}
/**********************************************************
* elementwise product of two real column vectors. 
**********************************************************/
static void product5(double *A, int *irA, int *jcA, 
                     double *B, int *irB, int *jcB, double *P, 
                     int n, int type, int isspA, int isspB)

{  int  k, kx, ky, kx2, ky2, rx, ry;

      if ( !isspA & !isspB ) {
         for (k=0; k<n; k++){ 
             P[k] = A[k]*B[k];
         }
      }
      else if ( isspA & !isspB ) {
         kx  = jcA[0];     
         kx2 = jcA[1]-1;
	 for (k=kx; k<=kx2; k++) {
             rx = irA[k]; 
             P[rx] = A[k]*B[rx];
         }   
      }
      else if ( !isspA & isspB ) {
         ky  = jcB[0];     
         ky2 = jcB[1]-1;
	 for (k=ky; k<=ky2; k++) {
             ry = irB[k]; 
             P[ry] = A[ry]*B[k];
         }   
      }
      else if ( isspA & isspB ) {
           kx = jcA[0];  kx2 = jcA[1]-1;  rx = irA[kx];
           ky = jcB[0];  ky2 = jcB[1]-1;  ry = irB[ky]; 
           while ((kx <= kx2) & (ky <= ky2)){
               if (rx == ry) { 
		  P[rx] = A[kx]*B[ky]; 
                  kx++; ky++; 
                  rx = irA[kx];
                  ry = irB[ky]; } 
               else if (rx < ry) { 
                  kx++; 
                  rx = irA[kx]; }
               else { 
                  ky++; 
                  ry = irB[ky]; }                   
	   }
      }
      return; 
}
/**********************************************************/
void mexFunction(
	int  nlhs,  mxArray        *plhs[],
	int  nrhs,  const mxArray  *prhs[] )
{
   mxArray  *A_cell_pr, *B_cell_pr, *blk_cell_pr, *lhs; 
   double   *A,   *AI,  *B,   *BI,  *P,   *PPI; 
   int      *irA, *jcA, *irB, *jcB, *irP, *jcP;
   double   *Ptmp, *PPItmp, *blksize;
   int      *cumsum, *blknnz; 
   int      mblk, nblk, mA, nA, mB, nB, m1, n1, m2, n2; 
   int      iscell_blk, iscell_A, iscell_B, iscmpA, iscmpB, isspA, isspB, index;
   int      type;

   int subs[2]; 
   int nsubs=2; 
   int numblk, cols, NZmax, idx, i, l, j, k, kstart, kend;

/* Check for proper number of arguments */
   if (nrhs<=3){
     mexErrMsgTxt("mexProd2: requires at least 3 input arguments."); }
   else if (nlhs>2){ 
     mexErrMsgTxt("mexProd2: requires 1 output argument."); }

/* get matrix dimension */
   
   iscell_blk = mxIsCell(prhs[0]); 
   mblk = mxGetM(prhs[0]); 
   nblk = mxGetN(prhs[0]); 
   iscell_A = mxIsCell(prhs[1]); 
   mA = mxGetM(prhs[1]);
   nA = mxGetN(prhs[1]); 
   iscell_B = mxIsCell(prhs[2]); 
   mB = mxGetM(prhs[2]); 
   nB = mxGetN(prhs[2]);
   if (!iscell_A) { mA = 1; nA = 1;}
   if (!iscell_B) { mB = 1; nB = 1;}
   if ((iscell_A & !iscell_B) || (!iscell_A & iscell_B)) {
      mexErrMsgTxt("mexProd2: 2ND and 3RD input must both be matrices or \
                    cell arrays");}   
   if (mA != mB || nA != 1 || nB != 1) { 
      mexErrMsgTxt("mexProd2: 2ND and 3RD input not compatible"); }   
   if (!iscell_blk) { 
      mexErrMsgTxt("mexProd2: 1ST input must be the cell array blk"); }   
   if (nblk != 2) { 
      mexErrMsgTxt("mexProd2: 1ST input must be a cell array with 2 columns");}
   if (mA != mblk || mB !=mblk) {
      mexErrMsgTxt("mexProd2: 1ST input has too many or too few rows"); }
   if (nrhs == 3) { type = 0; }
   else{ type = (int)*mxGetPr(prhs[3]); }

/*** Create a cell array for the return argument if A,B are cell arrays ***/
   
   if (iscell_A & iscell_B) { 
      plhs[0] = mxCreateCellMatrix(mA,1); 
   } 
/***********************************************/     
    for (i=0; i < mA; i++){
       if (iscell_A) {
          subs[0] = i; 
          subs[1] = 0;
          index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
          A_cell_pr = mxGetCell(prhs[1],index); 
          A  = mxGetPr(A_cell_pr); 
   	  m1 = mxGetM(A_cell_pr); 
          n1 = mxGetN(A_cell_pr); 
          isspA  = mxIsSparse(A_cell_pr);               
          iscmpA = mxIsComplex(A_cell_pr); 
          if (iscmpA) { AI = mxGetPi(A_cell_pr); }
          if (isspA) {
             irA = mxGetIr(A_cell_pr);
    	     jcA = mxGetJc(A_cell_pr); }   
       }
       else {
          A = mxGetPr(prhs[1]); 
          isspA = mxIsSparse(prhs[1]);
          iscmpA = mxIsComplex(prhs[1]); 
          m1 = mxGetM(prhs[1]); 
          n1 = mxGetN(prhs[1]); 
          if (iscmpA) { AI = mxGetPi(prhs[1]); }
          if (isspA) { irA = mxGetIr(prhs[1]);
	               jcA = mxGetJc(prhs[1]); }
       }
       if (iscell_B) {
          subs[0] = i; 
          subs[1] = 0;
          index = mxCalcSingleSubscript(prhs[2],nsubs,subs); 
          B_cell_pr = mxGetCell(prhs[2],index); 
          B  = mxGetPr(B_cell_pr); 
          m2 = mxGetM(B_cell_pr); 
          n2 = mxGetN(B_cell_pr); 
          isspB  = mxIsSparse(B_cell_pr);
          iscmpB = mxIsComplex(B_cell_pr);  
          if (iscmpB) { BI = mxGetPi(B_cell_pr); }
          if (isspB) {
             irB = mxGetIr(B_cell_pr);
	     jcB = mxGetJc(B_cell_pr); }
       }
       else {
          B = mxGetPr(prhs[2]); 
          isspB  = mxIsSparse(prhs[2]);
          iscmpB = mxIsComplex(prhs[2]);  
          m2 = mxGetM(prhs[2]); 
          n2 = mxGetN(prhs[2]); 
          if (iscmpB) { BI = mxGetPi(prhs[2]); }
          if (isspB) {
             irB = mxGetIr(prhs[2]);
             jcB = mxGetJc(prhs[2]); }
       }   
       if (m1 != m2 || n1 != n2) { 
           mexErrMsgTxt("mexProd2: 2ND and 3RD input not compatible"); }   
       subs[0] = i; 
       subs[1] = 1;
       index = mxCalcSingleSubscript(prhs[0],nsubs,subs); 
       blk_cell_pr = mxGetCell(prhs[0],index); 
       blksize = mxGetPr(blk_cell_pr);
       numblk  = mxGetN(blk_cell_pr); 
       /******************************
       *  create return argument 
       *******************************/          
       if (isspA & isspB & n1 > 1) { 
          cumsum = mxCalloc(numblk+1,sizeof(int)); 
          blknnz = mxCalloc(numblk+1,sizeof(int)); 
          cumsum[0] = 0; blknnz[0] = 0;
          for (k=0; k<numblk; k++) {
             cols = (int)blksize[k]; 
	     blknnz[k+1] = blknnz[k] + cols*cols;
             cumsum[k+1] = cumsum[k] + cols; 
	  }        
          Ptmp = mxCalloc(m1,sizeof(double)); 
          if (iscmpA || iscmpB) { PPItmp = mxCalloc(m1,sizeof(double)); }
          NZmax = blknnz[numblk];
          if (!iscell_A & !iscell_B) {  
             if (iscmpA || iscmpB) { 
	        plhs[0] = mxCreateSparse(m1,n1,NZmax,mxCOMPLEX); 
                PPI = mxGetPi(plhs[0]); }
             else {
 	        plhs[0] = mxCreateSparse(m1,n1,NZmax,mxREAL); }
             P   = mxGetPr(plhs[0]); 
             irP = mxGetIr(plhs[0]); 
             jcP = mxGetJc(plhs[0]); 
	  }
          else {
             if (iscmpA || iscmpB) { 
	        lhs = mxCreateSparse(m1,n1,NZmax,mxCOMPLEX); 
                PPI = mxGetPi(lhs); }
             else {
 	        lhs = mxCreateSparse(m1,n1,NZmax,mxREAL); }
             P   = mxGetPr(lhs); 
             irP = mxGetIr(lhs); 
             jcP = mxGetJc(lhs); 
          }
       }
       else {
          if (!iscell_A & !iscell_B) {  
             if (iscmpA || iscmpB) { 
                plhs[0] = mxCreateDoubleMatrix(m1,n1,mxCOMPLEX); 
                PPI = mxGetPi(plhs[0]); }
             else {
 	        plhs[0] = mxCreateDoubleMatrix(m1,n1,mxREAL); }
             P = mxGetPr(plhs[0]); }
          else {
             if (iscmpA || iscmpB) { 
	        lhs = mxCreateDoubleMatrix(m1,n1,mxCOMPLEX); 
                PPI = mxGetPi(lhs); }
             else {
 	        lhs = mxCreateDoubleMatrix(m1,n1,mxREAL); }
             P = mxGetPr(lhs); 
	  }
       }
       /**********************************************
        * Do the actual computations in a subroutine 
        **********************************************/
       if (m1==m2 & n1==n2 & m1==n1 & n1 > 1) {
          if ( !isspA & !isspB ){
             product(A, AI, B, BI, P, PPI, m1, type, iscmpA, iscmpB); }
          else if ( !isspA & isspB ){ 
             product2(A, AI, B, BI, irB, jcB, P, PPI, m1, type, \
                      iscmpA, iscmpB); }
          else if ( isspA & !isspB ){ 
             product3(A, AI, irA, jcA, B, BI, P, PPI, m1, type, \
                      iscmpA, iscmpB); }
          else if ( isspA & isspB ){
             product4(A, AI, irA, jcA, B, BI, irB, jcB, P, PPI, irP, jcP, \
                      Ptmp, PPItmp, cumsum, numblk, type, iscmpA, iscmpB); }
       }
       else if (m1==m2 & n1==n2 & n1==1) {
 	  if (iscmpA || iscmpB) {
	     mexErrMsgTxt("mexProd2: column vectors must be real."); }
          else {
             product5(A,irA,jcA,B,irB,jcB,P,m1,type,isspA,isspB); 
	  }
       }
       /**********************************
	* put result in the ith cell
        **********************************/
       if (iscell_A & iscell_B) {
          subs[0] = i; 
          subs[1] = 0;
          index = mxCalcSingleSubscript(plhs[0],nsubs,subs);
          mxSetCell(plhs[0],index,lhs); 
       }
    }
    return;
}
/**********************************************************/

