/******************************************************************
* mexProd2nz: compute the elements of P whose positions are
*             specified by the variable list.    
* 
*   P = mexProd2nz(blk,A,B,list,sym), where P = A'*B. 
*   
*   output: P is sparse, where 
*           P(i,j) = <ith column of A, jth column of B>
*
*   Important note: if sym = 1, only the upper triangular part
*                   of P is returned. 
*
*   The variable list must have its 2nd column sorted in ascending
*   order. The first and second column of list indicate the row 
*   and column indices of elements of P to be computed. 
*    
*   E.g. list = [1 1]
*               [2 2]
*               [3 2]
*               [2 3]
*   
* SDPT3: version 3.0 
* Copyright (c) 1997 by
* K.C. Toh, M.J. Todd, R.H. Tutuncu
* Last Modified: 2 Feb 01
******************************************************************/

#include <math.h>
#include <mex.h>

/**********************************************************
* realdot: 
**********************************************************/
double realdot(const double *x, const int idx1, const double *y, 
               const int idx2,  const int n)

{  int i;
   double r; 

   r=0.0;
   for(i=0; i< n-3; i++){                 /* LEVEL 4 */
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; 
   }
   if(i < n-1){                           /* LEVEL 2 */
      r+= x[i+idx1] * y[i+idx2]; i++;
      r+= x[i+idx1] * y[i+idx2]; i++;
   }
   if(i < n){                             /* LEVEL 1 */
      r+= x[i+idx1] * y[i+idx2]; 
   }
   return r; 
}
/**********************************************************
*  B dense
*  A = nxp, B = nxp
*  Q(i,j) = <ith column of A, jth column of B>
**********************************************************/
void  prod1(int m, int n, int p, 
            double *A,  int *irA, int *jcA, int isspA, 
            double *B, 
            double *Q,  int *irQ, int *jcQ, double *list, int len, int sym)

{ int     j, k, r, t, rn, jn, jm, jold, kstart, kend, idx, count;
  double  tmp;  

     jold = -1; count = 0; 
     for (t=0; t<len; ++t) {
         r = (int) list[t]-1;      /** subtract 1 to adjust for Matlab index **/
         j = (int) list[len+t]-1;  
         if (j != jold) { jn = j*n;  jm = j*m;  jold = j; }       
         if (!sym || (sym && r<=j)) {
            if (!isspA) {
               rn = r*n;
               tmp = realdot(A,rn,B,jn,n);  
               /** for (k=0; k<n; ++k){ tmp += A[k+rn]*B[k+jn]; }  **/
            }
            else {
               tmp = 0;  
               kstart = jcA[r];   
               kend   = jcA[r+1]; 
   	       for (k=kstart; k<kend; ++k) {
                  idx = irA[k]; 
	          tmp += A[k]*B[idx+jn]; } 
            }
            Q[count] = tmp; 
            irQ[count] = r;  jcQ[j+1]++;  count++; 
         }
     }
     for (k=0; k<p; k++) { jcQ[k+1] += jcQ[k]; } 
     return;
}
/**********************************************************
*  B sparse.
*  A = nxm, B = nxp  
*  Q(i,j) = <ith column of A, jth column of B>
**********************************************************/
void  prod2(int m, int n, int p,  
            double *A, int *irA, int *jcA, int isspA,
            double *B, int *irB, int *jcB, int isspB,
            double *Q, int *irQ, int *jcQ, double *Btmp, 
            double *list, int len, int sym)  

{ int     j, k, r, t, rn, jn, jm, jold, kstart, kend, idx, count;
  double  tmp;  

     jold = -1; count = 0;  
     for (t=0; t<len; ++t) {
         r = (int) list[t]-1;    /** subtract 1 to adjust for Matlab index **/
         j = (int) list[len+t]-1;  
         if (j != jold) {
            jn = j*n;  jm = j*m;  
            /***** copy j-th column of sparse B to Btmp *****/ 
            for (k=0; k<n; ++k) { Btmp[k] = 0; }
            kstart = jcB[j];  kend = jcB[j+1];
            for (k=kstart; k<kend; ++k) { idx = irB[k]; Btmp[idx] = B[k];  }
            jold = j; 
         }
         if (!sym || (sym && r<=j)) {
            tmp = 0;  
            if (!isspA) {
               rn = r*n; 
               tmp = realdot(A,rn,Btmp,0,n); 
               /** for (k=0; k<n; ++k){ tmp += A[k+rn]*Btmp[k]; }  **/
            }
            else {
               kstart = jcA[r];   kend = jcA[r+1]; 
               for (k=kstart; k<kend; ++k) {
                  idx = irA[k]; 
                  tmp += A[k]*Btmp[idx]; } 
            }
            Q[count] = tmp; 
            irQ[count] = r;  jcQ[j+1]++;   count++; 
	 }
     }
     for (k=0; k<p; k++) { jcQ[k+1] += jcQ[k]; } 
     return;
}
/**********************************************************
*  COMPLEX product
*  B dense
*  A = nxp, B = nxp
*  Q(i,j) = <ith column of A, jth column of B>
**********************************************************/
void  prod1cmp(int m, int n, int p, 
            double *A,  double *AI, int *irA, int *jcA, int isspA, 
            double *B,  double *BI,
            double *Q,  double *QI, int *irQ, int *jcQ, double *list, int len, int sym)

{ int     j, k, r, t, rn, jn, jm, jold, kstart, kend, idx, count;
  double  tmp, tmpI;  

     jold = -1; count = 0; 
     for (t=0; t<len; ++t) {
         r = (int) list[t]-1;      /** subtract 1 to adjust for Matlab index **/
         j = (int) list[len+t]-1;  
         if (j != jold) { jn = j*n;  jm = j*m;  jold = j; }       
         if (!sym || (sym && r<=j)) {
            if (!isspA) {
               rn = r*n;
               tmp  = realdot(A,rn,B,jn,n)  + realdot(AI,rn,BI,jn,n);  
               tmpI = realdot(A,rn,BI,jn,n) - realdot(AI,rn,B,jn,n); 
            }
            else {
 	       tmp = 0;  tmpI = 0; 
               kstart = jcA[r];   
               kend   = jcA[r+1]; 
   	       for (k=kstart; k<kend; ++k) {
                  idx = irA[k]+jn; 
	          tmp  += A[k]*B[idx]  + AI[k]*BI[idx]; 
                  tmpI += A[k]*BI[idx] - AI[k]*B[idx]; } 
            }
            Q[count] = tmp; QI[count] = tmpI; 
            irQ[count] = r;  jcQ[j+1]++;  count++; 
         }
     }
     for (k=0; k<p; k++) { jcQ[k+1] += jcQ[k]; } 
     return;
}
/**********************************************************
*  COMPLEX product
*  B sparse.
*  A = nxm, B = nxp  
*  Q(i,j) = <ith column of A, jth column of B>
**********************************************************/
void  prod2cmp(int m, int n, int p,  
            double *A, double *AI, int *irA, int *jcA, int isspA,
            double *B, double *BI, int *irB, int *jcB, int isspB,
	    double *Q, double *QI, int *irQ, int *jcQ, double *Btmp, double *BItmp,
            double *list, int len, int sym)  

{ int     j, k, r, t, rn, jn, jm, jold, kstart, kend, idx, count;
  double  tmp, tmpI;  

     jold = -1; count = 0;  
     for (t=0; t<len; ++t) {
         r = (int) list[t]-1;    /** subtract 1 to adjust for Matlab index **/
         j = (int) list[len+t]-1;  
         if (j != jold) {
            jn = j*n;  jm = j*m;  
            /***** copy j-th column of sparse B to Btmp *****/ 
            for (k=0; k<n; ++k) { Btmp[k] = 0; BItmp[k] = 0; }
            kstart = jcB[j];  kend = jcB[j+1];
            for (k=kstart; k<kend; ++k) { idx = irB[k]; Btmp[idx] = B[k]; BItmp[idx] = BI[k]; }
            jold = j; 
         }
         if (!sym || (sym && r<=j)) {
            if (!isspA) {
               rn = r*n; 
               tmp  = realdot(A,rn,Btmp,0,n)  + realdot(AI,rn,BItmp,0,n); 
               tmpI = realdot(A,rn,BItmp,0,n) - realdot(AI,rn,Btmp,0,n);  
            }
            else {
  	       tmp = 0; tmpI = 0;   
               kstart = jcA[r];   kend = jcA[r+1]; 
               for (k=kstart; k<kend; ++k) {
                  idx = irA[k]; 
                  tmp  += A[k]*Btmp[idx]  + AI[k]*BItmp[idx]; 
                  tmpI += A[k]*BItmp[idx] - AI[k]*Btmp[idx]; } 
            }
            Q[count] = tmp;  QI[count] = tmpI; 
            irQ[count] = r;  jcQ[j+1]++;   count++; 
	 }
     }
     for (k=0; k<p; k++) { jcQ[k+1] += jcQ[k]; } 
     return;
}
/**********************************************************
* 
**********************************************************/
void mexFunction(
	int  nlhs,  mxArray        *plhs[],
	int  nrhs,  const mxArray  *prhs[] )
{
   double   *A, *AI, *B, *BI, *Q, *QI; 
   int      *irA,  *jcA,  *irB,  *jcB, *irQ, *jcQ;
   double   *list, *Btmp, *BItmp; 
   int      sym, m1, n1, m2, n2, mlist, nlist, isspA, isspB, iscmpA, iscmpB;


/* Check for proper number of arguments */
   if (nrhs < 4){
     mexErrMsgTxt("mexProd2nz: requires at least 4 input arguments."); }
   else if (nlhs>2){ 
     mexErrMsgTxt("mexProd2nz: requires 1 output argument."); }
   
/***********************************************/     

       A = mxGetPr(prhs[1]); 
       isspA = mxIsSparse(prhs[1]);
       iscmpA = mxIsComplex(prhs[1]); 
       m1 = mxGetM(prhs[1]); 
       n1 = mxGetN(prhs[1]); 
       if (isspA) { irA = mxGetIr(prhs[1]);
                    jcA = mxGetJc(prhs[1]); }
       if (iscmpA) { AI = mxGetPi(prhs[1]); }
       B = mxGetPr(prhs[2]); 
       isspB  = mxIsSparse(prhs[2]);
       iscmpB = mxIsComplex(prhs[2]); 
       m2 = mxGetM(prhs[2]); 
       n2 = mxGetN(prhs[2]); 
       if (isspB) { irB = mxGetIr(prhs[2]);
                    jcB = mxGetJc(prhs[2]); }
       if (iscmpB) { BI = mxGetPi(prhs[2]); }  
       mlist = mxGetM(prhs[3]); 
       nlist = mxGetN(prhs[3]); 
       list  = mxGetPr(prhs[3]); 
       if (nrhs == 5) { sym = (int) *mxGetPr(prhs[4]); }
       else           { sym = 0; }
       if (m1!=m2) { 
          mexErrMsgTxt("mexProd2nz: A, B are not compatible"); } 
       if (nlist != 2 ) { 
          mexErrMsgTxt("mexProd2nz: the content of list must have 2 columns"); }           
       if (!iscmpA & !iscmpB) { 
	  plhs[0] = mxCreateSparse(n1,n2,mlist,mxREAL); }
       else {
          plhs[0] = mxCreateSparse(n1,n2,mlist,mxCOMPLEX); 
          QI = mxGetPi(plhs[0]); 
          if (!iscmpA) { 
	    if (!isspA) { AI = mxCalloc(m1*n1,sizeof(double)); }
            else        { AI = mxCalloc(jcA[n1],sizeof(double)); }
	  }
          if (!iscmpB) { 
	    if (!isspB) { BI = mxCalloc(m2*n2,sizeof(double)); }
            else        { BI = mxCalloc(jcB[n2],sizeof(double)); }
	  }
       }
       Q = mxGetPr(plhs[0]); irQ = mxGetIr(plhs[0]); jcQ = mxGetJc(plhs[0]);        

       /**********************************************
        * Do the actual computations in a subroutine 
        **********************************************/
        if (!iscmpA & !iscmpB) { 
           if (!isspB) { 
   	      prod1(n1,m2,n2,A,irA,jcA,isspA,B,Q,irQ,jcQ,list,mlist,sym); }
           else {
              Btmp = mxCalloc(m2,sizeof(double)); 
   	      prod2(n1,m2,n2,A,irA,jcA,isspA,B,irB,jcB,isspB,Q,irQ,jcQ,Btmp,list,mlist,sym); }
	}
        else { 
           if (!isspB) { 
   	      prod1cmp(n1,m2,n2,A,AI,irA,jcA,isspA,B,BI,Q,QI,irQ,jcQ,list,mlist,sym); }
           else {
              Btmp = mxCalloc(m2,sizeof(double)); 
              BItmp = mxCalloc(m2,sizeof(double)); 
   	      prod2cmp(n1,m2,n2,A,AI,irA,jcA,isspA,B,BI,irB,jcB,isspB,Q,QI,irQ,jcQ,Btmp,BItmp,list,mlist,sym); }
	}              
    return;
}		 
/**********************************************************/

