 /***********************************************************************
* mexschur.c : C mex file to compute  
*
*    schur(I,J) = schur(I,J) + Trace(Ai U Aj V),
*    where I = permA[i], J = permA[j]. 
*          
*    mexschur(blk,A,nzlistA,permA,U,V,type,schur);  
* 
*   input:  blk = 1x2 a cell array describing the block structure of A.
*           A   = 1xm CELL ARRAY where each cell is a real symmetric matrix,
*           nzlistA = 1xm CELL ARRAY where each cell is a matrix with 
*                     2 columns. For the matrix in the jth cell, its kth row
*                     records the row and column indices of the kth nonzero
*                     element in A_j. Note that the indices start from 1 
*                     in accordance with Matlab convention. 
*           permA = a permutation vector.  
*           U,V = either an 1x1 CELL ARRAY where each cell 
*                 is a real symmetric matrix, or a real symmetric matrix.
*           type = 0 corresponds to computing Trace(Ai*(U Aj V)),
*                = 1 corresponds to computing Trace(Ai*(U Aj V + V Aj U)/2)).
*
* SDPT3: version 2.1 
* Copyright (c) 1997 by
* K.C. Toh, M.J. Todd, R.H. Tutuncu
* Last Modified: 7 Jul 99   
************************************************************************/

#include <math.h>
#include <mex.h>
#include <time.h>

/**********************************************************
*  compute  Trace(B U*A*V)
*  A,B are assumed to be real and sparse.
*  U,V are assumed to be real and dense. 
**********************************************************/
double schurij(int n,  
             double *A, double *idxA, int nzA,
             double *B, double *idxB, int nzB, 
             double *U, double *V)

{ int    ra, rb, ca, cb, l, k; 
  double tmp, tmp0, tmp1, tmp2; 

  tmp0 = 0.0; tmp1 = 0.0; 
  for (l=0; l<nzB; ++l) {  
     rb = (int) idxB[l] -1;   
     cb = (int) idxB[l+nzB] -1; 
     tmp2 = 0.0; 
     if (rb <= cb) { 
        for (k=0; k<nzA; ++k) { 
           ra = (int) idxA[k] -1 ;
           ca = (int) idxA[k+nzA] -1; 
           tmp2 += U[rb+ra*n] * A[k] * V[ca+cb*n]; 
        }
        if (rb < cb) { tmp0 += B[l]*tmp2; }
        else         { tmp1 += B[l]*tmp2; }
     }
  }
  tmp = 2*tmp0 + tmp1;
  return tmp;
}
/**********************************************************
*  compute  Trace(B (U*A*V + V*A*U)/2 )
*  A,B are assumed to be real and sparse.
*  U,V are assumed to be real and dense. 
**********************************************************/
double schurij2(int n,  
             double *A, double *idxA, int nzA,
             double *B, double *idxB, int nzB, 
             double *U, double *V)

{ int    ra, rb, ca, cb, l, k, ind1, ind2; 
  double tmp, tmp0, tmp1, tmp2; 

  tmp0 = 0.0; tmp1 = 0.0; 
  for (l=0; l<nzB; ++l) {  
     rb = (int) idxB[l] -1;   
     cb = (int) idxB[l+nzB] -1; 
     tmp2 = 0.0; 
     if (rb <= cb) { 
        for (k=0; k<nzA; ++k) { 
           ra = (int) idxA[k] -1 ;
           ca = (int) idxA[k+nzA] -1; 
           ind1 = rb+ra*n; 
           ind2 = ca+cb*n;  
           tmp2 += (U[ind1]*V[ind2] + U[ind2]*V[ind1])*A[k]; 
        }
        if (rb < cb) { tmp0 += B[l]*tmp2; }
        else         { tmp1 += B[l]*tmp2; }
     }
  }
  tmp = tmp0 + tmp1/2;
  return tmp;
}
/**********************************************************
*  compute  Trace(B (U*A*V+V*A*U)/2) for sparse block 
*  diagonal matrices. 
*  A,B are assumed to be real and sparse.
*  U2, V2 are modified versions of U,V. 
**********************************************************/
double schurij3(int n,  
             double *A,  double *idxA, int nzA,
             double *B,  double *idxB, int nzB, 
             double *U2, double *V2,   double *blksize,
             int *cumsum, int *blknnz, int *col2blk, int type)

{ int     ra, rb, ca, cb, l, k, idx1, idx2, idx2cb;
  int     colstart, colend, cols, blkno; 
  int     colstart2, colend2, cols2, blkno2; 
  double  tmp, tmp0, tmp1, tmp2; 

  tmp0 = 0.0; tmp1 = 0.0; 
  for (l=0; l<nzB; ++l) {  
     rb = (int) idxB[l] -1;   
     cb = (int) idxB[l+nzB] -1; 
     tmp2 = 0.0; 
     if (rb <= cb) { 
        blkno2 = col2blk[cb]; 
        colstart2 = cumsum[blkno2]; 
        colend2   = cumsum[blkno2+1]; 
        cols2  = (int)blksize[blkno2];
        idx2cb =  blknnz[blkno2] + (cb-colstart2)*cols2;
        for (k=0; k<nzA; ++k) { 
           ra = (int) idxA[k] -1 ;
           ca = (int) idxA[k+nzA] -1; 
           if (ca >= colstart2 && ca < colend2) {
             /*** find (ca,cb) element of V from V2 ***/  
              idx2 = idx2cb + (ca-colstart2); 
             /*** find (rb,ra) element of U from U2 ***/ 
              blkno = col2blk[ra]; 
              colstart = cumsum[blkno]; 
              colend   = cumsum[blkno+1]; 
              cols  = (int)blksize[blkno]; 
              idx1 = blknnz[blkno] + (ra-colstart)*cols + (rb-colstart);
              if (type == 0) { 
                  tmp2 += U2[idx1] * A[k] * V2[idx2]; }
              else {
                  tmp2 += (U2[idx1]*V2[idx2] + U2[idx2]*V2[idx1])*A[k]; }
	   }
	}
        if (rb < cb) { tmp0 += B[l]*tmp2; }
        else         { tmp1 += B[l]*tmp2; }
     }
  }
  if (type == 0) {
     tmp = 2*tmp0 + tmp1; }
  else {
     tmp = tmp0 + tmp1/2; 
  }
  return tmp;
}
/**********************************************************/
void mexFunction(
     int nlhs,   mxArray  *plhs[], 
     int nrhs,   const mxArray  *prhs[] )
{    
     mxArray  *A_cell_pr, *B_cell_pr, *idxA_cell_pr, *idxB_cell_pr; 
     mxArray  *U_cell_pr, *V_cell_pr, *blk_cell_pr;

     double   *A, *B, *idxA, *idxB, *U, *V, *schur;
     double   *blksize, *dpermA; 
     int      *irU, *jcU, *irV, *jcV, *permA; 

     int      iscell_blk, iscell_A, iscell_idxA, iscell_U, iscell_V; 
     int      isspA, isspB, isspU, isspV; 

     double  *U2, *V2;
     int     *cumsum, *blknnz, *col2blk; 
     int     NZmax, cols, idx, r, l, colstart, colend, kstart, kend; 

     int     subs[2], dim[1];
     int     nsubs=2; 
     int     mA, nA, midxA, nidxA, mpermA, npermA, mU, nU, mV, nV, nzA, nzB; 
     int     type, m, index, idx1, idx2, n, i, j, k, numblk;
     double  tmp; 

/* CHECK THE DIMENSIONS */

   iscell_blk = mxIsCell(prhs[0]);
   
   iscell_A = mxIsCell(prhs[1]); 
   mA = mxGetM(prhs[1]); 
   nA = mxGetN(prhs[1]); 

   iscell_idxA = mxIsCell(prhs[2]);
   midxA = mxGetM(prhs[2]); 
   nidxA = mxGetN(prhs[2]); 

   dpermA = mxGetPr(prhs[3]); 
   mpermA = mxGetM(prhs[3]); 
   npermA = mxGetN(prhs[3]); 
   permA  = mxCalloc(npermA,sizeof(int)); 
   for (i=0; i<npermA; ++i) { permA[i] = (int)dpermA[i]-1; } 

   iscell_U = mxIsCell(prhs[4]); 
   mU = mxGetM(prhs[4]);
   nU = mxGetN(prhs[4]);

   iscell_V = mxIsCell(prhs[5]); 
   mV = mxGetM(prhs[5]);
   nV = mxGetN(prhs[5]);

   if (nrhs < 7) {
      mexErrMsgTxt(" mexschur: must have at least 7 inputs"); }
   if (nrhs == 7) { type = 0; }
   else           { type = (int) *mxGetPr(prhs[7]); }
  
   if (nlhs > 1) { 
      mexErrMsgTxt("mexschur: requires no output argument"); }
   if (!iscell_A || !iscell_idxA) {
      mexErrMsgTxt(" mexschur: A and splistA must be a cell array "); }
   if (mA > 1 || midxA > 1) { 
      mexErrMsgTxt(" mexschur: A and splistA must be a cell array with \
                     1 row only"); }
   if (!iscell_U) { mU = 1; nU = 1; }
   if (!iscell_V) { mV = 1; nV = 1; }
   if ((!iscell_U && iscell_V) || (iscell_U && !iscell_V)) {
     mexErrMsgTxt("mexschur: both U,V must be cell arrays or matrices"); }
   if (!iscell_blk) {
     mexErrMsgTxt("mexschur: 1ST input must be the cell array blk"); }

   subs[0] = 0; 
   subs[1] = 1;
   index = mxCalcSingleSubscript(prhs[0],nsubs,subs); 
   blk_cell_pr = mxGetCell(prhs[0],index); 
   blksize = mxGetPr(blk_cell_pr); 
   numblk = mxGetN(blk_cell_pr);
   dim[0] = 0; for (i=0; i<numblk; ++i) { dim[0] += (int)blksize[i]; }
   n = dim[0]; 

/**** get pointers for schur matrix ****/

   schur = mxGetPr(prhs[6]); 
   m = mxGetM(prhs[6]);

/**** get pointers for U,V ****/    

   if (iscell_U) {
      subs[0] = 0; 
      subs[1] = 0;
      index = mxCalcSingleSubscript(prhs[4],nsubs,subs); 
      U_cell_pr = mxGetCell(prhs[4],index); 
      U  = mxGetPr(U_cell_pr); 
      mU = mxGetM(U_cell_pr); 
      nU = mxGetN(U_cell_pr); 
      isspU  = mxIsSparse(U_cell_pr);
      if (isspU) {
         irU = mxGetIr(U_cell_pr);
	 jcU = mxGetJc(U_cell_pr); }
   }
   else {
      U = mxGetPr(prhs[4]); 
      isspU  = mxIsSparse(prhs[4]);
      mU = mxGetM(prhs[4]); 
      nU = mxGetN(prhs[4]); 
      if (isspU) {
         irU = mxGetIr(prhs[4]);
         jcU = mxGetJc(prhs[4]); }
   }
   if (iscell_V) {
      subs[0] = 0; 
      subs[1] = 0;
      index = mxCalcSingleSubscript(prhs[5],nsubs,subs); 
      V_cell_pr = mxGetCell(prhs[5],index); 
      V  = mxGetPr(V_cell_pr); 
      mV = mxGetM(V_cell_pr); 
      nV = mxGetN(V_cell_pr); 
      isspV  = mxIsSparse(V_cell_pr);
      if (isspV) {
         irV = mxGetIr(V_cell_pr);
	 jcV = mxGetJc(V_cell_pr); }
   }
   else {
      V = mxGetPr(prhs[5]); 
      isspV  = mxIsSparse(prhs[5]);
      mV = mxGetM(prhs[5]); 
      nV = mxGetN(prhs[5]); 
      if (isspV) {
         irV = mxGetIr(prhs[5]);
         jcV = mxGetJc(prhs[5]); }
    }
    if (nU == 1 || nV == 1) {
       mexErrMsgTxt("mexschur: matrices must be square, with dimension > 1"); }
    if (nU != n || nV != n) { 
       mexErrMsgTxt("mexschur: size of U or V is not compatible with blk "); }
    if (numblk == 1 && (isspU || isspV)) { 
       mexErrMsgTxt("mexschur: numblk = 1, matrices U, V must be dense"); }
    if (numblk > 1 && (!isspU || !isspV)) { 
       mexErrMsgTxt("mexschur: numblk > 1, matrices U, V must be sparse"); }

/******************************
 * copy U to U2, V to V2
 ******************************/

    if (numblk >1 && isspU && isspV) {
       cumsum = mxCalloc(numblk+1,sizeof(int)); 
       blknnz = mxCalloc(numblk+1,sizeof(int)); 
       col2blk = mxCalloc(n,sizeof(int)); 
       blknnz[0]=0; cumsum[0]=0; 
       for (k=0; k<numblk; ++k){
          cols = (int)blksize[k];
          blknnz[k+1] = blknnz[k]+cols*cols; 
          cumsum[k+1] = cumsum[k]+cols; }
       NZmax = blknnz[numblk];
       U2 = mxCalloc(NZmax,sizeof(double)); 
       V2 = mxCalloc(NZmax,sizeof(double)); 
       for (l=0; l<numblk; ++l){
           cols = (int)blksize[l]; 
           colstart=cumsum[l]; colend = cumsum[l+1]; 
           for (j=colstart; j<colend; ++j){
               col2blk[j] = l; 
               idx = blknnz[l] + (j-colstart)*cols;
               kstart = jcU[j];
               kend   = jcU[j+1];
               for (k=kstart; k<kend; ++k){
                   r = irU[k]-colstart; 
                   U2[idx+r] = U[k]; }
               kstart = jcV[j];
               kend   = jcV[j+1];
               for (k=kstart; k<kend; ++k){
                   r = irV[k]-colstart; 
                   V2[idx+r] = V[k]; }
	    }
	}
    }
   
/************************************
 * compute schur(i,j)
 ***********************************/
      for (j=0; j<nidxA; j++){
          subs[0] = 0; 
          subs[1] = j;
          index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
          A_cell_pr = mxGetCell(prhs[1],index); 
          A  = mxGetPr(A_cell_pr); 
          isspA  = mxIsSparse(A_cell_pr);               
          if (!isspA) { 
             mexErrMsgTxt("mexschur: each matrix in A must be sparse"); }
          index = mxCalcSingleSubscript(prhs[2],nsubs,subs);
          idxA_cell_pr = mxGetCell(prhs[2],index); 
          idxA = mxGetPr(idxA_cell_pr);
          nzA = mxGetM(idxA_cell_pr); 
          if (nzA > 0) { 
             for (i=0; i<=j; i++){
                subs[0] = 0; 
                subs[1] = i;
                index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
                B_cell_pr = mxGetCell(prhs[1],index); 
                B = mxGetPr(B_cell_pr); 
                index = mxCalcSingleSubscript(prhs[2],nsubs,subs);
                idxB_cell_pr = mxGetCell(prhs[2],index); 
                idxB =  mxGetPr(idxB_cell_pr);
                nzB = mxGetM(idxB_cell_pr);
                if (nzB > 0) {   
                   if (numblk == 1) {                 
                      if (type == 0) {  
                         tmp = schurij(n,B,idxB,nzB,A,idxA,nzA,U,V); }
                      else if (type == 1) { 
                         tmp = schurij2(n,B,idxB,nzB,A,idxA,nzA,U,V); }  }
                   else {
                      tmp = schurij3(n,B,idxB,nzB,A,idxA,nzA,U2,V2, \
                                     blksize,cumsum,blknnz,col2blk,type);
		   }
                   idx1 = permA[i] + permA[j]*m;
                   idx2 = permA[j] + permA[i]*m; 
                   schur[idx1] += tmp; 
                   schur[idx2] = schur[idx1];   
		}
	     }
	  }
      }
   return;
   }
/**********************************************************/

