/***********************************************************************
* mextrace.c : C mex file to compute the trace of two matrices, 
*              Trace(A*B'). 
*          
*   Tr = mextrace(blk,A,B,type);  
* 
*   input:  blk = a cell array describing the block structure of A and B.
*           A = either an pxm  CELL ARRAY where each cell 
*                      is a real symmetric matrix,
*               or a real symmetric matrix.
*           B = either an px1 CELL ARRAY  where each cell 
*                      is a real symmetric matrix, 
*               or a real symmetric matrix.
*        type = 0, matrices A,B may be non-hermitian
*             = 1, matrices A,B are hermitian
*
*  output:  Tr = mx1 vector, where
*           Tr(j) = sum_{i=1}^p TRACE(A{i,j}*B{i,1}'). 
*
* SDPT3: version 2.1 
* Copyright (c) 1997 by
* K.C. Toh, M.J. Todd, R.H. Tutuncu
* Last Modified: 7 Jul 99   
************************************************************************/

#include <math.h>
#include <mex.h>
#include <time.h>

static int max1, max2; 
#define IMAX(a,b) (max1=(a),max2=(b),(max1)>(max2)?(max1):(max2))

/**********************************************************
* compute Trace(A*B), 
*  type = 0, A,B are general real matrices. 
*  type = 1, A,B are both real symmetric matrices
***********************************************************/
double trace(int n,
             int isspX, double *x, int *xrows, int *xcols,
             int isspY, double *y, int *yrows, int *ycols, int type) 
{     
   double  u, d, Tr;
   int     i, j, k, r, jn, idx, kstart, kend, kx, ky, kx2, ky2, rx, ry;

   u = 0; 
   d = 0; 
   if (type == 1) {  
      if ( !isspX && !isspY ) {
         for (j=0; j<n; j++){
             jn = j*n;
             for (i=0; i<j; i++){
                 idx = i+jn; 
                 u += x[idx]*y[idx]; }
             idx = j+jn; 
             d += x[idx]*y[idx];
         }
      }
      else if ( isspX && !isspY ) {   
         for (j=0; j<n; j++){
             kstart = xcols[j]; 
             kend   = xcols[j+1];
             jn = j*n;  r = -1;
             for (k=kstart; k<kend; k++) {
                 r = xrows[k]; 
                 if (r >= j) { break; } 
	         u += x[k]*y[r+jn]; 
             }
             if (r == j) { d += x[k]*y[r+jn]; }
         }
      }
      else if ( !isspX && isspY ) {   
         for (j=0; j<n; j++){
             kstart = ycols[j]; 
             kend   = ycols[j+1];
             jn = j*n; r = -1;
             for (k=kstart; k<kend; k++) {
                r = yrows[k]; 
                if (r >= j) { break; }     
	        u += y[k]*x[r+jn]; 
             }
             if (r == j) { d += y[k]*x[r+jn]; }
         }
      }
      else if ( isspX && isspY ) {   
         for (j=0; j<n; j++){
             ky  = ycols[j]; 
             ky2 = ycols[j+1];
             if (ky < ky2) { 
                kx  = xcols[j]; 
                kx2 = xcols[j+1]; 
                ry  = yrows[ky];    
                for (k=kx; k<kx2; ++k) {
                   rx = xrows[k];
                   while (ry < rx && ky < ky2) { 
                      ky++;  
                      ry  = yrows[ky]; }
                   if (rx > j) { break; }                                 
                   if (rx == ry) {
                      if (rx < j)  
                         { u += x[k]*y[ky]; }   
                      else
                         { d += x[k]*y[ky]; }                     
		   }
		}
	     }
         }
      }
      Tr = 2*u+d;
   }
   else if (type == 0) {
      if ( !isspX && !isspY ) {
         for (j=0; j<n; j++){
             jn = j*n;
             for (i=0; i<n; i++){
                 idx = i + jn; 
                 u += x[idx]*y[idx]; }
         }
      }
      else if ( isspX && !isspY ) {   
         for (j=0; j<n; j++){
             kstart = xcols[j]; 
             kend   = xcols[j+1];
             jn = j*n;
             for (k=kstart; k<kend; ++k){
                r = xrows[k];               
	        u += x[k]*y[r+jn]; 
             }
         }
      }
      else if ( !isspX && isspY ) {   
         for (j=0; j<n; j++){
             kstart = ycols[j]; 
             kend   = ycols[j+1];
             jn = j*n;  
             for (k=kstart; k<kend; ++k){
                r = yrows[k]; 
  	        u += y[k]*x[r+jn]; 
             }
         }
      }
      else if ( isspX && isspY ) {   
         for (j=0; j<n; j++){
             kx = xcols[j]; 
             ky = ycols[j]; 
             rx = xrows[kx];
             ry = yrows[ky]; 
             kx2 = xcols[j+1];
             ky2 = ycols[j+1];
             while ((kx < kx2) && (ky < ky2)){
                 if (rx == ry) { 
 		    u += x[kx]*y[ky]; 
                    kx++; ky++; 
                    rx = xrows[kx];
                    ry = yrows[ky]; } 
                 else if (rx < ry) { 
                    kx++; 
                    rx = xrows[kx]; }
                 else {  
                    ky++; 
                    ry = yrows[ky]; }                   
             }
         }
      }
      Tr = u;
   }
   return Tr;
}
/*********************************************
* compute the real part of Trace(A*B'), 
* type = 0, A,B are general complex matrices.
*      = 1, A,B are complex hermitian matrices
**********************************************/
double cmptrace(int n,
          int isspX, double *x, double *xm, int *xrows, int *xcols,
          int isspY, double *y, double *ym, int *yrows, int *ycols, int type) 
{     
   double  u, d, Tr;
   int     i, j, k, r, jn, idx, kx, ky, kx2, ky2, rx, ry, kstart, kend;

   u = 0; 
   d = 0; 
   if (type == 1) {
      if ( !isspX && !isspY ) {
         for (j=0; j<n; j++){
             jn = j*n;
             for (i=0; i<j; i++){
                 idx = i + jn; 
                 u += x[idx]*y[idx] + xm[idx]*ym[idx]; }
             idx = j+jn; 
             d += x[idx]*y[idx] + xm[idx]*ym[idx];
         }
      }
      else if ( isspX && !isspY ) {   
         for (j=0; j<n; j++){
             kstart = xcols[j]; 
             kend = xcols[j+1];
             jn = j*n;  r = -1;
             for (k=kstart; k<kend; ++k) {
                r = xrows[k]; 
                idx = r+jn; 
                if (r >= j) { break; } 
                u += x[k]*y[idx] + xm[k]*ym[idx]; 
	     }
             if (r == j) {d += x[k]*y[idx] + xm[k]*ym[idx]; }
         }
      }
      else if ( !isspX && isspY ) {   
         for (j=0; j<n; j++){
             kstart = ycols[j]; 
             kend = ycols[j+1];
             jn = j*n;  r = -1;
             for (k=kstart; k<kend; ++k) {
               r = yrows[k]; 
               idx = r+jn; 
               if (r >=j ) { break; }
               u += y[k]*x[idx] + ym[k]*xm[idx];
	     }
             if (r == j) {d += y[k]*x[idx] + ym[k]*xm[idx]; }
         }
      }
      else if ( isspX && isspY ) {   
         for (j=0; j<n; j++){
             kx = xcols[j]; 
             ky = ycols[j]; 
             rx = xrows[kx];
             ry = yrows[ky]; 
             kx2 = xcols[j+1];
             ky2 = ycols[j+1];
             while ((kx < kx2) && (ky < ky2)){
                 if (rx == ry) { 
                    if (rx == j) { 
                       d += x[kx]*y[ky] + xm[kx]*ym[ky]; 
                       break; }
 		    if (rx <j) { 
                       u += x[kx]*y[ky] + xm[kx]*ym[ky]; }
                    else { break; }
                    kx++; ky++; 
                    rx = xrows[kx];
                    ry = yrows[ky]; } 
                 else if (rx < ry) { 
                    if (rx >= j) { break;}  
                    kx++; 
                    rx = xrows[kx]; }
                 else {
                    if (ry >= j) { break;}                      
                    ky++; 
                    ry = yrows[ky]; }
             }
         }
      }
      Tr = 2*u+d;
   }
   else if (type == 0) {
      if ( !isspX && !isspY ) {
         for (j=0; j<n; j++){
             jn = j*n;
             for (i=0; i<n; i++){
                 idx = i + jn; 
                 u += x[idx]*y[idx] + xm[idx]*ym[idx]; }
         }
      }
      else if ( isspX && !isspY ) {   
         for (j=0; j<n; j++){
             kstart = xcols[j]; 
             kend = xcols[j+1];
             jn = j*n;
             for (k=kstart; k<kend; ++k){
                r = xrows[k]; 
                idx = r+jn; 
	        u += x[k]*y[idx] + xm[k]*ym[idx]; 
             }
         }
      }
      else if ( !isspX && isspY ) {   
         for (j=0; j<n; j++){
             kstart = ycols[j]; 
             kend = ycols[j+1]; 
             jn = j*n;
             for (k=kstart; k<kend; ++k){
               r = yrows[k]; 
               idx = r+jn; 
	       u += y[k]*x[idx] + ym[k]*xm[idx]; 
             }
         }
      }
      else if ( isspX && isspY ) {   
         for (j=0; j<n; j++){
             kx = xcols[j]; 
             ky = ycols[j]; 
             rx = xrows[kx];
             ry = yrows[ky]; 
             kx2 = xcols[j+1];
             ky2 = ycols[j+1];
             while ((kx < kx2) && (ky < ky2)){
                 if (rx == ry) { 
 		    u += x[kx]*y[ky] + xm[kx]*ym[ky]; 
                    kx++; ky++; 
                    rx = xrows[kx];
                    ry = yrows[ky]; } 
                 else if (rx < ry) { 
                    kx++; 
                    rx = xrows[kx]; }
                 else { 
                    ky++; 
                    ry = yrows[ky]; }                   
             }
         }
      }
      Tr = u;
   }
   return Tr;
}
/**********************************************************
* compute inner product of x and y, where x,y are real vectors. 
**********************************************************/
double innerprod(int n, 
                 int isspX, double *x, int *xrows, int *xcols,
                 int isspY, double *y, int *yrows, int *ycols) 

{     double Tr; 
      int    k, kx, ky, kx2, ky2, rx, ry;

      Tr = 0.0; 
      if ( !isspX && !isspY ) {
         for (k=0; k<n; k++){ 
             Tr += x[k]*y[k];
         }
      }
      else if ( isspX && !isspY ) {
         kx  = xcols[0];     
         kx2 = xcols[1];
	 for (k=kx; k<kx2; k++) {
             rx = xrows[k]; 
             Tr += x[k]*y[rx];
         }   
      }
      else if ( !isspX && isspY ) {
         ky  = ycols[0];     
         ky2 = ycols[1];
	 for (k=ky; k<ky2; k++) {
             ry = yrows[k]; 
             Tr += y[k]*x[ry];
         }   
      }
      else if ( isspX && isspY ) {
           kx = xcols[0]; 
           ky = ycols[0]; 
           rx = xrows[kx];
           ry = yrows[ky]; 
           kx2 = xcols[1];
           ky2 = ycols[1];
           while ((kx < kx2) && (ky < ky2)){
                 if (rx == ry) { 
		    Tr += x[kx]*y[ky]; 
                    kx++; ky++; 
                    rx = xrows[kx];
                    ry = yrows[ky]; } 
                 else if (rx < ry) { 
                    kx++; 
                    rx = xrows[kx]; }
                 else { 
                    ky++; 
                    ry = yrows[ky]; }                   
	   }
      }
      return Tr; 
}
/**********************************************************
* compute trace of sparse matrices with block structure
* specified by blksize. 
**********************************************************/
double sptrace(int *cumsum, int* blknnz, int numblk,
               double *A, double *AI, int *irA,  int *jcA, 
               double *Btmp, double *BItmp, int type, int iscmpA, int iscmpB) 

{ double  u, d, Tr; 
  int idx, l, j, k, r, cols, row, colstart, colend, kstart, kend;

  u = 0; d = 0;
  if (type == 1) {
     for (l=0; l<numblk; ++l){
        colstart=cumsum[l];
        colend  =cumsum[l+1];
        cols = colend-colstart;
        for (j=colstart; j<colend; j++){
           kstart = jcA[j]; 
           kend = jcA[j+1]; 
           idx = blknnz[l] + (j-colstart)*cols; 
           row = -1;
           if (!iscmpA || !iscmpB) { 
              for (k=kstart; k<kend; ++k){	      
                 row = irA[k];
                 r = idx + row-colstart;
                 if (row >= j) { break; }
                 u += Btmp[r]*A[k]; 
	      } 
              if (row == j)  { d += Btmp[r]*A[k]; }
	   }         
           else { 
              for (k=kstart; k<kend; ++k){	      
	         row = irA[k];
                 r = idx + row-colstart;
                 if (row >= j) { break; }
                 u += Btmp[r]*A[k] + BItmp[r]*AI[k]; 
	      }
	      if (row == j)  { d += Btmp[r]*A[k] + BItmp[r]*AI[k]; } 
	   }         
        }
     }
     Tr = d+2*u; 
  }
  else if (type == 0) {
     for (l=0; l<numblk; ++l){
        colstart=cumsum[l];
        colend  =cumsum[l+1];
        cols = colend-colstart;
        for (j=colstart; j<colend; j++){
           kstart = jcA[j]; 
           kend = jcA[j+1]; 
           idx = blknnz[l] + (j-colstart)*cols; 
           if  (!iscmpA || !iscmpB) {
              for (k=kstart; k<kend; ++k){
	         r = idx + irA[k]-colstart; 
                 u += Btmp[r]*A[k]; }  
	   }
           else {
              for (k=kstart; k<kend; ++k){
	         r = idx + irA[k]-colstart; 
                 u += Btmp[r]*A[k] + BItmp[r]*AI[k]; }
	   }
        }
     }
     Tr = u; 
  }
  return Tr; 
}
/**********************************************************/
void mexFunction(
      int nlhs,   mxArray  *plhs[], 
      int nrhs,   const mxArray  *prhs[] )
{    
     mxArray  *A_cell_pr, *B_cell_pr, *blk_cell_pr;
     double   *A, *B, *AI, *BI, *TR;
     int      *irA, *jcA, *irB, *jcB; 
     double   *Btmp, *BItmp, *blksize;
     int      *cumsum, *blknnz; 
     int      type, isspA, isspB, iscell_blk, iscell_A, iscell_B, iscmpA, iscmpB;
     int      NZmax; 

     double  Tr;
     int     subs[2];
     int     nsubs=2; 
     int     mblk, nblk, mA, nA, mB, nB, m1, n1, m2, n2, i, j, k, index;    
     int     idx, l, r, cols, colstart, colend, kstart, kend, numblk;


/* CHECK THE DIMENSIONS */

   iscell_blk = mxIsCell(prhs[0]);
   mblk = mxGetM(prhs[0]); 
   nblk = mxGetN(prhs[0]);
   iscell_A = mxIsCell(prhs[1]); 
   mA = mxGetM(prhs[1]); 
   nA = mxGetN(prhs[1]); 
   iscell_B = mxIsCell(prhs[2]); 
   mB = mxGetM(prhs[2]);
   nB = mxGetN(prhs[2]);

   if (nrhs < 3) {
      mexErrMsgTxt(" mextrace: must have at least 3 inputs"); }
   else if (nrhs == 3) {
      type = 1;  }
   else {
      type = (int)*mxGetPr(prhs[3]); }   
   if (nlhs>2) { 
      mexErrMsgTxt("mextrace: requires 1 output argument"); }

   if (!iscell_A) { mA = 1; nA = 1; }
   if (!iscell_B) { mB = 1; nB = 1; }
   if (!iscell_A && iscell_B){
     mexErrMsgTxt("mextrace: A is a not cell array, but B is a cell array"); }
   if ( iscell_B && nB > 1){
     mexErrMsgTxt("mextrace: cell array B can only have 1 column"); }
   if (!iscell_blk) {
     mexErrMsgTxt("mextrace: 1ST input must be the cell array blk"); }
   if (mblk != mA || mblk != mB) {
     mexErrMsgTxt("mextrace: 1ST input has too many or too few rows"); }

/**** Create a matrix for the return argument and assign pointers ****/

   plhs[0] = mxCreateDoubleMatrix(nA,1,mxREAL); 
   TR = mxGetPr(plhs[0]); 
   
/* Do the computations in a subroutine */
    
    for (i=0; i < mB; i++){
       if (iscell_B) {
          subs[0] = i; 
          subs[1] = 0;
          index = mxCalcSingleSubscript(prhs[2],nsubs,subs); 
          B_cell_pr = mxGetCell(prhs[2],index); 
          B  = mxGetPr(B_cell_pr); 
          m2 = mxGetM(B_cell_pr); 
          n2 = mxGetN(B_cell_pr); 
          isspB  = mxIsSparse(B_cell_pr);
          iscmpB = mxIsComplex(B_cell_pr);  
          if (iscmpB) { BI = mxGetPi(B_cell_pr); }
          if (isspB) {
             irB = mxGetIr(B_cell_pr);
	     jcB = mxGetJc(B_cell_pr); }
       }
       else {
          B = mxGetPr(prhs[2]); 
          isspB  = mxIsSparse(prhs[2]);
          iscmpB = mxIsComplex(prhs[2]);  
          m2 = mxGetM(prhs[2]); 
          n2 = mxGetN(prhs[2]); 
          if (iscmpB) { BI = mxGetPi(prhs[2]); }
          if (isspB) {
             irB = mxGetIr(prhs[2]);
             jcB = mxGetJc(prhs[2]); }
       }
       subs[0] = i; 
       subs[1] = 1;
       index = mxCalcSingleSubscript(prhs[0],nsubs,subs); 
       blk_cell_pr = mxGetCell(prhs[0],index); 
       blksize = mxGetPr(blk_cell_pr); 
       numblk = mxGetN(blk_cell_pr);
       /******************************
	* copy B to Btmp, BI to BItmp
	******************************/
       if (numblk > 1 && isspB) {
          cumsum = mxCalloc(numblk+1,sizeof(int)); 
          blknnz = mxCalloc(numblk+1,sizeof(int)); 
          blknnz[0]=0; cumsum[0]=0; 
          for (k=0; k<numblk; ++k){
             cols = (int)blksize[k];
             blknnz[k+1] = blknnz[k]+cols*cols; 
             cumsum[k+1] = cumsum[k]+cols; }
          NZmax = blknnz[numblk];
          Btmp = mxCalloc(NZmax,sizeof(double)); 
          if (iscmpB) { BItmp = mxCalloc(NZmax,sizeof(double)); }
          for (l=0; l<numblk; ++l){
             colstart=cumsum[l]; colend = cumsum[l+1]; 
             cols = (int)blksize[l]; 
             for (j=colstart; j<colend; ++j){
                kstart = jcB[j];
                kend   = jcB[j+1];
                idx = blknnz[l] + (j-colstart)*cols;
                for (k=kstart; k<kend; ++k){
                   r = irB[k]-colstart; 
                   Btmp[idx+r] = B[k]; }
                if (iscmpB) {
                   for (k=kstart; k<kend; ++k){
                      r = irB[k]-colstart; 
                      BItmp[idx+r] = BI[k]; }
		}
	     }
	  }
       }
       /************************************
	* compute trace(Aj*B')
	***********************************/
       for (j=0; j < nA; j++){
	  if (iscell_A) {
             subs[0] = i; 
             subs[1] = j;
             index = mxCalcSingleSubscript(prhs[1],nsubs,subs); 
             A_cell_pr = mxGetCell(prhs[1],index); 
             A  = mxGetPr(A_cell_pr); 
             m1 = mxGetM(A_cell_pr); 
             n1 = mxGetN(A_cell_pr);
             isspA  = mxIsSparse(A_cell_pr);               
             iscmpA = mxIsComplex(A_cell_pr); 
             if (iscmpA) { AI = mxGetPi(A_cell_pr); }
             if (isspA) {
                irA = mxGetIr(A_cell_pr);
    	        jcA = mxGetJc(A_cell_pr); }   }
          else {
             A = mxGetPr(prhs[1]); 
             m1 = mxGetM(prhs[1]); 
             n1 = mxGetN(prhs[1]);
             isspA = mxIsSparse(prhs[1]);
             iscmpA = mxIsComplex(prhs[1]); 
             if (iscmpA) { AI = mxGetPi(prhs[1]); }
             if (isspA) { irA = mxGetIr(prhs[1]);
	                  jcA = mxGetJc(prhs[1]); }
          }
          if (numblk>1 && isspA && isspB) {
             Tr = sptrace(cumsum,blknnz,numblk,A,AI,irA,jcA,Btmp,BItmp, \
                  type,iscmpA,iscmpB);  }
          else if (m1==m2 && m2==n2) {
  	     if (iscmpA && iscmpB) {
                Tr = cmptrace(m2,isspA,A,AI,irA,jcA,isspB,B,BI,irB,jcB,type);}
             else { 
	        Tr = trace(m2,isspA,A,irA,jcA,isspB,B,irB,jcB,type); } 
	  }
          else if (m1==m2 && n2==1) {
	       Tr = innerprod(m2,isspA,A,irA,jcA,isspB,B,irB,jcB); } 
          else { mexErrMsgTxt("mextrace: size of matrices are different"); }
          TR[j] += Tr;
       }
    }
 return;
}
/**********************************************************/

