%%***************************************************************
%% nonzerolist: find the combined list of non-zero elements
%%     of Aj, j = 1:k, for each k,  
%%     assuming that the Aj's are permuted such 
%%     that A1 has the fewest nonzero elements,
%%     followed by A2, and so on. If the Aj's are 
%%     not already permuted in the sense just described,
%%     then permutation is carried out in this program
%%     when finding the combined list of non-zero 
%%     elements.
%%             
%% [dcumlistA,splistA,permA,listRd] = nonzerolist(blk,A,spdensity,C); 
%%
%% Suppose the dimension of the cell array A is pxm. 
%%
%%  spdensity = the density threshold where dcumlistA{p,k} is set to
%%              'full'.
%%
%%  dcumlistA = pxm cell array. The (p,k) cell, dcumlistA{p,k}, is a 
%%          matrix with 2 columns where each row is the 
%%          row and column index of the nonzero elements
%%          of PERMUTED A{p,k} that are not already present
%%          in the combined list from permuted A{p,1} through A{p,k-1}.
%%          Note that if the number of nonzero elements exceeds  
%%          a prescribed threshold, then dcumlistA{p,k} = 'full'.
%%  splistA = pxM cell array. The (p,k) cell, splistA{p,k}, is a matrix
%%         with 2 columns where each row is the row and column index 
%%         of the nonzero elements of the PERMUTED A{p,k}.
%%         M here denotes the number of matrices in A1,...,Am that 
%%         are very sparse. 
%%  permA = pxm matrix, where each row permA(i,:) describes 
%%          the permutation such that the PERMUTED
%%          A{i,1}, A{i,2},...,A{i,m} has number of nonzero 
%%          elements arranged in ascending order. 
%%
%% SDPT3: version 2.1
%% Copyright (c) 1997 by
%% K.C. Toh, M.J. Todd, R.H. Tutuncu
%% Last Modified: 7 Jul 99
%%**************************************************************

  function [dcumlistA,splistA,permA,listRd] = nonzerolist(blk,A,spdensity,C); 

    if (nargin <= 3); C = []; end;
    if (nargin <= 2); spdensity = 0.3; end; 
%%
    listRd = cell(size(blk,1)); 
    for p = 1:size(blk,1);
       blktmp = blk{p,2}; 
       n = sum(blktmp);
       nztol = spdensity*sum(blktmp.*blktmp); 
       if strcmp(blk{p,1},'nondiag'); 
          for k = 1:size(A,2); 
              dcumlistA{p,k} = 'full';
              splistA{p,k} = 'full'; 
              nnzA(p,k) = nnz(A{p,k}); 
          end; 
          [dummy,permA(p,:)] = sort(nnzA(p,:)); 
          Acum = sparse(n,n); 
          numnz = 0; 
          breakyes = 0;
          for k = 1:size(A,2);
             ind = permA(p,k); 
             Aind = A{p,ind};              
             [I,J] = find(Aind); 
             if (length(I) == 0); 
                dcumlistA{p,k} = [];
             elseif (length(I) <= nztol); 
                tmp = zeros(length(I),1); 
                for t = 1:length(I); 
                   i = I(t); j = J(t); 
                   tmp(t) = Acum(i,j);           
                end;
                %% find new nonzero positions when 
                %% Aind is added to Acum. 
                idx = find(tmp == 0);   
                numnz = numnz + length(idx); 
                if (numnz <= nztol); 
                   dcumlistA{p,k} = [I(idx) J(idx)];
                else
                   %% the rest are all 'full'
                   breakyes = 1; 
                end;
             else
                %% the rest are all 'full'
                breakyes = 1; 
             end;
             Acum = Acum + inf*abs(Aind); 
             if ~isstr(dcumlistA{p,k})
                listRd{p} = [listRd{p}; dcumlistA{p,k}];
             else
                listRd{p} = 'full';  
             end
             %%
             %% nonzero list of very sparse A's. 
             %%
             if (length(I) == 0); 
                splistA{p,k} = []; 
                breakyes = 0; 
             elseif (length(I) <= spdensity*n)
                splistA{p,k} = [I J];  
                breakyes = 0; 
             end;  
             if breakyes; break; end; 
          end
          if (~isempty(C) & ~isstr(listRd{p})); 
             [I,J] = find(C{p});
             if (length(I) > 0); 
                tmp = zeros(length(I),1); 
                for t = 1:length(I); 
                   i = I(t); j = J(t); 
                   tmp(t) = Acum(i,j);           
                end;
                idx = find(tmp == 0);   
                numnz = numnz + length(idx); 
                if (numnz <= nztol); 
                   listC = [I(idx) J(idx)];
                else
                   listC = 'full'; 
                end;
                if ~isstr(listC); 
                   listRd{p} = [listRd{p}; listC]; 
                else
                   listRd{p} = 'full';
                end
             end;
          end
       elseif strcmp(blk{p,1},'diag'); 
          permA(p,:) = [1:size(A,2)]; 
          for k = 1:size(A,2); 
             dcumlistA{p,k} = 'full';
             splistA{p,k} = 'full';  
          end;
          listRd{p} = 'full';
       end
    end
%%*******************************************************




