%%*****************************************************************************
%% preprocess - preprocess input data to detect infeasibility,
%%              eliminate dependencies, etc.
%%
%% [A,b,y0,indeprows,feasible] = preprocess(blk,A,b,y0,isspA);
%%
%% Input: A = a cell array with A{k} = A_k, k = 1:m, 
%%         C,b: data for the SDP instance.
%%
%%  Output:  A = a cell array with A{k} = A_k, k = 1:m, 
%%         C,b: data for the SDP instance.
%%         feasible : feasibility flag
%% 
%% SDPT3: version 2.1 
%% Copyright (c) 1997 by
%% K.C. Toh, M.J. Todd, R.H. Tutuncu
%% Last modified: 7 Jul 99
%%*****************************************************************************

   function [A,b,y0,indeprows,feasible] = preprocess(blk,A,b,y0,isspA);

   feasible=1;
%%
%%-----------------------------------------
%% delete zero matrices
%%-----------------------------------------
%%
   m = size(A,2); 
   normA = zeros(m,1);
   for k = 1:m; 
      Ak = A(:,k); 
      normAk = 0; 
      for p = 1:length(Ak); 
          normAk = normAk + norm(Ak{p},'fro');          
      end;
      normA(k) = normAk;  
   end;
   if any(normA == 0)
      zerorows=find(normA == 0);
      if any(b(zerorows) ~= 0)
         fprintf('\n preprocessor: problem infeasible\n');
         fprintf('Zero row(s) with nonzero rhs\n');
         fprintf(1,'%c',7);  % ring a bell
         feasible = 0; return;
      end;
      nonzerorows=find(normA > 0);
      A  = A(:,nonzerorows);
      b  = b(nonzerorows);
      y0 = y0(nonzerorows); 
      fprintf('\n preprocessor: %i zero-matrices deleted\n', ...
                   length(zerorows));
      m=length(b);
   end;
%%
%%-----------------------------------------
%% eliminate dependent matrices 
%%-----------------------------------------
%%  
    N = 0; 
    for p = 1:size(blk,1); 
       if strcmp(blk{p,1},'nondiag'); 
          N = N + sum(blk{p,2}.*blk{p,2}); 
       elseif strcmp(blk{p,1},'diag');
          N = N + blk{p,2};
       end;
    end;
    if ~any(isspA); 
       vecA = zeros(m,N);
    else; 
       vecA = sparse(m,N); 
    end;
    for k = 1:m; 
       vecAk = []; 
       for p = 1:size(blk,1)
          blktmp=blk{p,2};
          s = [0 cumsum(blktmp)];
          for i = 1:length(blktmp); 
              pos = [s(i)+1 : s(i+1)];
              if strcmp(blk{p,1},'nondiag'); 
                 B = A{p,k}(pos,pos);  
              elseif strcmp(blk{p,1},'diag');
                 B = A{p,k}(pos);  
              end;
              vecAk = [vecAk B(:)'];      
           end;
        end;
        vecA(k,:) = vecAk; 
    end;
%%
%% row reduce [ vecA | b ] to remove dependent rows/detect infeasibility
%%
   [indeprows,feasible]=rowreduce(vecA,b);
   if feasible
      A  = A(:,indeprows);
      b  = b(indeprows);
      y0 = y0(indeprows); 
      m  = length(b);
   else
      fprintf('\n preprocessor: problem infeasible\n');
      fprintf('Inconsistent constraints\n');
      fprintf(1,'%c',7);  % ring a bell
      return;
   end;
%%
%%*****************************************************************************
%% rowreduce -- eliminates dependent rows/ detects inconsistencies
%%
%% adapted from the MATLAB m file rref.m
%%
%%*****************************************************************************
%% SDPT3: version 2.1 
%% Copyright (c) 1998 by
%% K.C. Toh, M.J. Todd, R.H. Tutuncu
%% Last modified: July 1998
%%*****************************************************************************

   function [indeprows,feasible] = rowreduce(A,b,tol);
%%
   [m,n] = size(A);
   mn=min(m,n);
%% Compute the default tolerance if none was provided.
   if (nargin < 3), tol = max(m,n)*eps*norm(A,'inf'); end

   PsLs=speye(m); % matrix of permutations and lower tri. elementary operations
   P=1:m;         % vector of permutations
   indeprows=P; 
%%
%% Loop over the m x min(m,n) (sub)matrix.
%%
   i = 1;
   j = 1;
   jb = [];
   while (i <= m) & (j <= mn)  
      %% Find value and index of largest element in the remainder of column j.
      [p,k] = max(abs(A(i:m,j))); k = k+i-1;
      if (p <= tol)
         % The column is negligible, zero it out.
         A(i:m,j) = zeros(m-i+1,1);
         j = j + 1;
      else
         jb = [jb j];
         %% Swap i-th and k-th rows.
         %%
         PsLs([i k],:) = PsLs([k i],:);
         P([i k]) = P([k i]);
         A([i k],j:mn) = A([k i],j:mn);
      %%
      %% Subtract multiples of the first m columns of 
      %% pivot row from all the higher indexed rows
      %%
         nzidxA=find(A(i,j:mn))+j-1;
         nzidxP=find(PsLs(i,:));
         for k = [i+1:m];
            if abs(A(k,j)) > tol
               PsLs(k,nzidxP) = PsLs(k,nzidxP)-(A(k,j)/A(i,j))*PsLs(i,nzidxP);
               A(k,nzidxA) = A(k,nzidxA) - (A(k,j)/A(i,j))*A(i,nzidxA);
            end;      
         end;
         i = i + 1;
         j = j + 1;
      end;
   end;
%%
%% now the rest of the columns (if any and if necessary)
%%
   while (j<=n) & length(jb)<mn
      % first calculate the lower part of "reduced" column
        A(i:m,j)=PsLs(i:m,:)*A(:,j);
        [p,k] = max(abs(A(i:m,j))); k = k+i-1;
        if (p <= tol)
      	   % The column is negligible, move on.
           j = j + 1;
      	else
	   % Remember column index
  	   jb = [jb j];
	   % Swap i-th and k-th rows.
	   % 
           if k~= i
              A([i k],j) = A([k i],j);
	      PsLs([i k],:) = PsLs([k i],:);
              P([i k]) = P([k i]);
           end;
           nzidxP=find(PsLs(i,:));
           for k = [i+1:m]
	      if abs(A(k,j)) > tol
	         PsLs(k,nzidxP)= ...
                 PsLs(k,nzidxP)-(A(k,j)/A(i,j))*PsLs(i,nzidxP);
	      end      
	   end
	   i = i + 1;
	   j = j + 1;
	end;
    end;
%%
   feasible=1;
   if i<m+1
      b=PsLs*b;
      if any(abs(b(i:m))>tol)
         feasible=0;
      else
         fprintf('\n preprocessor: %d dependant constraint(s) removed.\n', ...
                    m-i+1);
         indeprows=sort(P(1:i-1));
      end;
   end;
   return; 
%%=============================================================================
