%%*****************************************************************************
%% sdp: solve an SDP by infeasible path-following method. 
%%
%%  vers = 1:  AHO  
%%  vers = 2:  HKM
%%  vers = 3:  NT   
%%  vers = 4:  GT
%%
%%  functions:  AHOpred, AHOcorr    
%%              HKMpred, HKMcorr 
%%              NTpred,  NTcorr
%%              GTpred,  GTcorr
%%              validate   [validate SDP data]  
%%              scaling    [scale SDP data]  
%%              blkchol    [Cholesky factorizations]
%%              corrprim   [correct primal infeasibility]  
%%              steplength [calculate maximum feasible steplength]
%%
%%  [obj,X,y,Z,gaphist,infeashist,info,Xiter,yiter,Ziter] = ...
%%                                sdp(blk,A,C,b,X0,y0,Z0,OPTIONS);
%%
%%  Input: blk: a cell array describing the block diagonal structure
%%              of SDP data.
%%           A: a cell array with A{k} = A_k, k = 1:m, 
%%         C,b: data for the SDP instance.
%%  (X0,y0,Z0): an initial iterate.
%%     OPTIONS: a structure that specifies parameters required in sdp.m,
%%              (if it is not given, the default in parameters.m is used). 
%%
%%  Output: obj  = [blktrace(blk,C*X) b'*y].
%%          (X,y,Z): an approximately optimal solution or a primal or dual
%%                   infeasibility certificate. 
%%          gaphist = history of reduction in <X,Z>. 
%%          infeashist = history of reduction in the infeasibility measure. 
%%          info = [termination-code iters gap infeasibity CPUtime].
%%          (Xiter,yiter,Ziter): last iterates.
%%
%%----------------------------------------------------------------------------
%%  The OPTIONS structure specifies the required parameters: 
%%      vers  gam  predcorr  expon  gaptol  inftol  steptol  maxit 
%%      sw2PC_tol  use_corrprim     printyes        scale_data
%%      (all have default values set in parameters.m).
%%
%%  vers = version of the direction to use. 
%%  gam  = steplength parameter
%%         [default: gam = 0, it is adaptively selected within the algorithm]. 
%%  predcorr = 1, if use Mehrotra-type predictor-corrector, else = 0
%%         [default 1]. 
%%  expon = exponent in decrease of centering parameter sigma
%%         [default: [3 1 1 2], for AHO, HKM, NT, and GT respectively]. 
%%  gaptol = tolerance for duality gap as a fraction of the value of the 
%%         objective functions   
%%         [default 1e-8]
%%  inftol = tolerance for stopping due to suspicion of infeasibility
%%         [default 1e-13]
%%  steptol = tolerance for stopping due to small steps 
%%         [default 1e-6].
%%  maxit = maximum number of iteration allowed
%%         [default 50].
%%  sw2PC_tol = use PC rule only if infeasibility < sw2PC_tol
%%         [default inf].
%%  use_corrprim = 1, if correct primal infeasibility in each step 
%%         (after feasibility attained), else = 0 
%%         [default 1].
%%  printyes = 1, if want to display result in each iteration, else = 0
%%         [default 1].
%%  scale_data = 1, if want to scale the data before solving the problem, 
%%         else 0
%%         [default 0].
%%  schurfun = [], if no user supplied routine to compute the Schur matrix, 
%%             or, a string containing the file name where the Schur matrix
%%             is computed, 
%%             e.g.  schurfun = 'mcpHKMsch(X,Zinv)', or 
%%                   schurfun = 'thetaHKMsch(X,Zinv,Rd,schurfun_parms)'.  
%%  schurfun_parms = [], if no user supplied routine to compute the Schur 
%%             matrix, or, a cell array containing the parameters need in 
%%             the user supplied Schur routine. 
%%  randnstate = state to be used for matlab 5 random normal generator
%%         [default 0]. 
%%
%%*****************************************************************************
%% SDPT3: version 2.1
%% Copyright (c) 1997 by
%% K.C. Toh, M.J. Todd, R.H. Tutuncu
%% Last modified: 7 Jul 99
%%*****************************************************************************

  function [obj,X,y,Z,gaphist,infeashist,info,Xiter,yiter,Ziter] = sdp(blk,A,C,b,X0,y0,Z0,OPTIONS);
%%                                      
%%-----------------------------------------
%% get parameters from the OPTIONS structure. 
%%-----------------------------------------
%%
   if nargin <= 7; parameters; end; 
   warning off; 

   vers         = OPTIONS.vers; 
   gam          = OPTIONS.gam; 
   predcorr     = OPTIONS.predcorr; 
   expon        = OPTIONS.expon;
   gaptol       = OPTIONS.gaptol;
   inftol       = OPTIONS.inftol;
   steptol      = OPTIONS.steptol;
   maxit        = OPTIONS.maxit;
   sw2PC_tol    = OPTIONS.sw2PC_tol;
   use_corrprim = OPTIONS.use_corrprim;
   printyes     = OPTIONS.printyes;
   scale_data   = OPTIONS.scale_data;
   schurfun        = OPTIONS.schurfun; 
   schurfun_parms  = OPTIONS.schurfun_parms; 
   randnstate      = OPTIONS.randnstate; 
   spdensity = 0.3;
%%
   if isempty(gam) | isempty(predcorr) | isempty(use_corrprim) | ...
      ~all([expon sw2PC_tol gaptol steptol]);
      fprintf('\n The variable OPTIONS in ** parameters.m **');
      fprintf(' require default settings. \n'); 
      termcode = -10;
      break; 
   end;
   if all(vers-[1 2 3 4]); error('*** vers must be 1,2,3 or 4***'); end; 
   pri_infeas_tol = 1e-12; 
%%
%%-----------------------------------------
%% validate SDP data  
%%-----------------------------------------
%%
   if (nargin <= 4) | (isempty(X0) | isempty(y0) | isempty(Z0)); 
      [X0,y0,Z0] = infeaspt(blk,A,C,b); 
   end
   X = X0; y = y0; Z = Z0; 
   clear X0 y0 Z0; 
   if ~iscell(C); tmp = C; clear C; C{1} = tmp; end;
   if ~iscell(X); tmp = X; clear X; X{1} = tmp; end;
   if ~iscell(Z); tmp = Z; clear Z; Z{1} = tmp; end;
   clear tmp;
   [A,C,X,Z,isspA,iscmpA] = validate(blk,A,C,b,X,y,Z,spdensity);
   m = length(b);  n = ops(C,'getM');
%%
%%-----------------------------------------
%% scale SDP data 
%%-----------------------------------------
%%
   normb0 = max(1,norm(b)); normC0 = max(1,ops(C,'norm')); 
   if (scale_data == 1); 
      [A,C,b,X,y,Z,normA,normC,normb] = scaling(blk,A,C,b,X,y,Z);
   else
      normA = ones(m,1); normC = 1; normb = 1;
   end;
%%
%%-----------------------------------------
%% compute A*A' to determine if the matrices
%% Ak are linearly independent. If they are
%% dependent, preprocess to get a linearly
%% independent set.  
%%-----------------------------------------
%%
   AAt = zeros(m); 
   for k = [1:m];  
       tmp = blktrace(blk,A(:,1:k),A(:,k));  
       AAt(1:k,k) = tmp;  AAt(k,1:k) = tmp'; 
   end;
   if (nnz(AAt) <= spdensity*m*m); AAt = sparse(AAt); end;
   permUA = symrcm(AAt); 
   [UA,indef] = chol(AAt(permUA,permUA));
   if indef | (min(abs(diag(UA))) < 1e-6); 
      fprintf('\n matrices Ak are nearly linearly dependent.'); 
      [A,b,y,indeprows,feasible] = preprocess(blk,A,b,y,isspA);
      if use_corrprim; 
         AAt = AAt(indeprows,indeprows); 
         permUA = symrcm(AAt); 
         [UA,indef] = chol(AAt(permUA,permUA));
      end; 
   end; 
   if (~use_corrprim); clear UA  permUA; UA = []; permUA = []; end; 
   clear AAt;
%%
%%-----------------------------------------
%% find the combined list of non-zero
%% elements of Aj, j = 1:k, for each k. 
%% ** must be done after the last 3 steps. 
%%-----------------------------------------
%%
   [listA,splistA,permA,listRd] = nonzerolist(blk,A,spdensity);
%%
%%-----------------------------------------
%% decomposition of A_k 
%%-----------------------------------------
%% 
   if (vers == 1 | vers == 4); 
      for p = 1:size(blk,1) 
         if (isspA(p) == 0); [LA(p,:),TA(p,:),PA(p,:)] = aasen(A(p,:)); end;
      end; 
   elseif (vers == 3); 
      for p = 1:size(blk,1) 
         if (isspA(p) == 0); AR(p,:) = Atriu(A(p,:)); end;
      end; 
   end;
%%
%%-----------------------------------------
%% initialization
%%-----------------------------------------
%%
   rp = b-blktrace(blk,A,X); 
   Rd = ops(C,'-',ops(Z,'+',Asum(blk,A,y,iscmpA))); 
   obj = (normb*normC)*[blktrace(blk,C,X)  b'*y];
   trXZ = blktrace(blk,X,Z); 
   gap = (normb*normC)*trXZ; 
   mu  = trXZ/n;  
   rel_gap = gap/max(1,mean(abs(obj)));
   normb2 = max(1,norm(b)); normC2 = max(1,ops(C,'norm')); 
   pri_infeas  = norm(rp)/normb2;
   dual_infeas = ops(Rd,'norm')/normC2;
 %%pri_infeas  = normb*norm(normA.*rp)/normb0; %% for original data
 %%dual_infeas = normC*ops(Rd,'norm')/normC0;  %% for original data
   infeas_meas = max(pri_infeas,dual_infeas); 
   pstep = 0; dstep = 0; pred_convg_rate = 1; corr_convg_rate = 1;
   termcode = -6; 
   gaphist(1) = gap; 
   infeashist(1) = infeas_meas;
   maxblk = ops(blk(:,2),'max'); 
   pri_infeas_bad = 0;
   indefsch = 0;   
%%
%%-----------------------------------------
%% display parameters, and initial info
%%-----------------------------------------
%%
   fprintf('\n********************************************');
   fprintf('***********************\n');
   fprintf('         Infeasible path-following algorithms'); 
   if ~isempty(schurfun);  
      fprintf('\n\nuser supplied Schur routine: '); 
      fprintf(schurfun); 
   end;
   fprintf('\n********************************************');
   fprintf('***********************\n');
   fprintf(' version  predcorr  gam  expon  use_corrprim  sw2PC_tol  scale_data\n');
   fprintf('   %1.0f        %1.0f      %4.3f',vers,predcorr,gam);
   fprintf('   %1.0f         %1.0f          %2.1e          %1.0f\n',...
           expon(vers),use_corrprim,sw2PC_tol,scale_data); 
   if printyes;
      fprintf('\nit  pstep dstep p_infeas d_infeas  gap')
      fprintf('        obj       sigma\n');
      fprintf('------------------------------------------------');
      fprintf('-------------------\n');
      fprintf('%2.0f  %4.3f %4.3f',0,0,0);
      fprintf(' %2.1e %2.1e  %2.1e',pri_infeas,dual_infeas,gap);
      fprintf(' %- 7.6e\n',mean(obj));
   end;
%%
%%---------------------------------------------------------------
%% start main loop
%%---------------------------------------------------------------
%%
   choltime = 0; predtime = 0; step_predtime = 0;
   corrtime = 0; step_corrtime = 0; misctime = 0; 
   for iter = 1:maxit;  
       flops(0); 
       time = zeros(1,7); 
       tstart = cputime;
       time(1) = cputime;
       [Xchol,indef(1)] = blkchol(blk,X); 
       [Zchol,indef(2)] = blkchol(blk,Z); 
       invXchol = cell(size(X));  %% must initialize.
       invZchol = cell(size(Z));           
       time(2) = cputime;
       choltime_iter = time(2)-time(1); 
       choltime = choltime + choltime_iter;
       if any(indef);  
          fprintf('\n Stop: X, Z are not both positive definite');
          termcode = -3;
          break; 
       end; 
%%
%%---------------------------------------------------------------
%% predictor step.
%%---------------------------------------------------------------
%%
     if predcorr & (infeas_meas < sw2PC_tol); 
        sigma = 0; 
     else; 
        sigma = 1-0.9*min(pstep,dstep); 
        if (iter == 1); sigma = 0.5; end; 
     end;
%%
%%-----------------------------------------
%% determine which way to compute the Schur 
%% matrix from the CPU time taken via two
%% different ways in the 3nd and 4th iteration
%% when the data correspond to the pth block 
%% is dense. 
%%-----------------------------------------
%%  
     if isempty(schurfun); 
        schur_opt = 2*ones(1,size(blk,1)); 
        dense_idx = find(isspA == 0); 
        if ~isempty(dense_idx); 
           if (iter > 1); schur_cpu_hist(iter-1,:) = schur_cpu; end;
           if ~all(iter-[1 3]); 
              schur_opt(dense_idx) = ones(1,length(dense_idx));  
           elseif ~all(iter-[2 4]); 
              schur_opt(dense_idx) = 2*ones(1,length(dense_idx)); 
           else; 
             tmp = schur_cpu_hist(3,dense_idx)./(schur_cpu_hist(4,dense_idx)+.01);
             schur_opt(dense_idx) = (tmp > 1) + 1; 
           end;
        end;
     else;
        schur_opt = schurfun; 
     end;
%%
     if (vers == 1);    
        [dX,dy,dZ,hRd,V,d,B,hX,Lsch,Usch,Psch,schur_cpu] = ...
            AHOpred(blk,A,rp,Rd,sigma,mu,schur_opt,schurfun_parms,...
                    iscmpA,normA,listA,permA,listRd,X,Z,LA,TA,PA); 
        invXchol = ops(Xchol,'inv');  clear Xchol;
        invZchol = ops(Zchol,'inv');  clear Zchol; 
     elseif (vers == 2);
        invZchol = ops(Zchol,'inv');  clear Zchol; 
        [dX,dy,dZ,hRd,Zinv,Uchol,indefsch,schur_cpu] = ...
            HKMpred(blk,A,rp,Rd,sigma,mu,schur_opt,schurfun_parms,iscmpA,...
                    normA,listA,permA,listRd,splistA,X,Xchol,invZchol);
        invXchol = ops(Xchol,'inv');  clear Xchol;
     elseif (vers == 3);
        invZchol = ops(Zchol,'inv'); 
        [dX,dy,dZ,hRd,W,G,sv,Uchol,indefsch,schur_cpu] = ...
            NTpred(blk,A,rp,Rd,sigma,mu,schur_opt,schurfun_parms,iscmpA,...
                   normA,listA,permA,listRd,splistA,X,Zchol,invZchol,AR);
        clear Zchol; 
        invXchol = ops(Xchol,'inv');  clear Xchol;
     elseif (vers == 4);
        invZchol = ops(Zchol,'inv'); 
        [dX,dy,dZ,hRd,ZH,invZH,d,sv2,phi2,psi2,Uchol,indefsch,schur_cpu] = ...
            GTpred(blk,A,rp,Rd,sigma,mu,schur_opt,schurfun_parms,...
                   iscmpA,normA,listA,permA,listRd,X,Zchol,invZchol,LA,TA,PA); 
        clear Zchol; 
        invXchol = ops(Xchol,'inv');  clear Xchol;
     end;
     if indefsch; 
        fprintf('\n Stop: Cholesky factorization of Schur matrix fails. '); 
        termcode = -4; 
        break; 
     end;
     if any(dy == inf);
        disp('   Schur complement system is singular');  
        CPUtime(iter) = cputime-tstart; 
        termcode = -4;
        break;
     end;
     time(3) = cputime;
     predtime_iter = time(3)-time(2);
     predtime = predtime + predtime_iter; 
%%
%%-----------------------------------------
%% step-lengths for predictor step
%%-----------------------------------------
%%
     if (gam == 0); 
        gamused = 0.9 + 0.08*min(pstep,dstep); 
     else;
        gamused = gam;
     end; 
     if use_corrprim & (pri_infeas < pri_infeas_tol)
        dX = corrprim(blk,A,dX,UA,permUA); end; 
     Xstep = steplength(blk,invXchol,dX,randnstate); 
     Zstep = steplength(blk,invZchol,dZ,randnstate); 
     pstep = min(1,gamused*Xstep); 
     dstep = min(1,gamused*Zstep); 
     trXZpred = trXZ + pstep*blktrace(blk,dX,Z) + dstep*blktrace(blk,X,dZ) ...
                + pstep*dstep*blktrace(blk,dX,dZ); 
     gappred = (normb*normC)*trXZpred; 
     mupred  = trXZpred/n; 
     mupredhist(iter) = mupred; 
     step_pred = min(pstep,dstep);
     time(4) = cputime; 
     step_predtime = step_predtime + (time(4)-time(3)); 
%%
%%-----------------------------------------
%%  stopping criteria for predictor step.
%%-----------------------------------------
%%
     if (pstep < steptol) | (dstep < steptol);
        fprintf('\n Stop: steps in predictor too short:');
        fprintf(' pstep = %3.2e,  dstep = %3.2e\n',pstep,dstep);
        CPUtime(iter) = cputime-tstart; 
        termcode = -2; 
        break; 
     end;
     %% test for slow progress when mu is small
     %%
     if (iter >= 2); 
        idx = [max(2,iter-5) : iter];
        pred_convg_rate = mean(mupredhist(idx)./mupredhist(idx-1));
        idx = [max(2,iter-2) : iter];
        pred_slow = all(mupredhist(idx)./mupredhist(idx-1) > 0.2);
     end; 
     if (~predcorr) | (predcorr & (infeas_meas >= sw2PC_tol)); 
        if (mu < 1e-8 & infeas_meas < 1e-8) ...
           & ((mupred/mu > 5*pred_convg_rate) | (pred_slow));
           fprintf('\n Stop: lack of progress in predictor:');
           fprintf(' mupred/mu = %3.2f, pred_convg_rate = %3.2f\n',...
                     mupred/mu,pred_convg_rate);
           CPUtime(iter) = cputime-tstart; 
           termcode = -1; 
           break;
        else; 
           X = ops(X,'+',dX,pstep);  y = y+dstep*dy;  
           Z = ops(Z,'+',dZ,dstep);
           gap = gappred;  mu = mupred;
        end;
     end;
%%
%%---------------------------------------------------------------
%%  corrector step.
%%---------------------------------------------------------------
%%
    if predcorr & (infeas_meas < sw2PC_tol) & ~indefsch;

       Xpred = ops(X,'+',dX,pstep);  ypred = y+dstep*dy;  
       Zpred = ops(Z,'+',dZ,dstep);
       mupred = blktrace(blk,Xpred,Zpred)/n; 
%%
       if (mu > 1e-6);
          if (step_pred < 1/sqrt(3)); 
             expon_used = 1; 
          else
             expon_used = max(expon(vers),3*step_pred^2); 
          end;
       else; 
          expon_used = max(1,min(expon(vers),3*step_pred^2)); 
       end; 
       sigma = min( 1, (mupred/mu)^expon_used ); 
%%
       if (vers == 1); 
          [dX,dy,dZ] = AHOcorr(blk,dX,dZ,A,rp,Rd,sigma,mu,hRd,...
                       Lsch,Usch,Psch,iscmpA,schur_opt,X,Z,V,d,B,hX); 
          clear Lsch Usch Psch V B hX; 
       elseif (vers == 2); 
          [dX,dy,dZ] = HKMcorr(blk,dX,dZ,A,rp,Rd,sigma,mu,hRd,...
                               Uchol,iscmpA,X,Zinv);
          clear Uchol Zinv; 
       elseif (vers == 3);
          [dX,dy,dZ] = NTcorr(blk,dX,dZ,A,rp,Rd,sigma,mu,hRd,...
                              Uchol,iscmpA,W,G,sv); 
          clear Uchol W G; 
       elseif (vers == 4); 
          [dX,dy,dZ] = GTcorr(blk,dX,dZ,A,rp,Rd,sigma,mu,hRd,...
                 Uchol,iscmpA,ZH,invZH,d,sv2,phi2,psi2);
          clear Uchol ZH invZH; 
       end;
       time(5) = cputime;
       corrtime_iter = time(5)-time(4);
       corrtime = corrtime + corrtime_iter; 
%%
%%-----------------------------------
%% step-lengths for corrector step
%%-----------------------------------
%%
       if (gam == 0); 
          gamused = 0.9 + 0.09*min(pstep,dstep); 
       else;
          gamused = gam;
       end;  
       if use_corrprim & (pri_infeas < pri_infeas_tol)
          dX = corrprim(blk,A,dX,UA,permUA); end; 
       Xstep = steplength(blk,invXchol,dX,randnstate); 
       Zstep = steplength(blk,invZchol,dZ,randnstate); 
       pstep = min(1,gamused*Xstep); 
       dstep = min(1,gamused*Zstep); 
%%
       trXZcorr = trXZ + pstep*blktrace(blk,dX,Z) + dstep*blktrace(blk,X,dZ)...
                  + pstep*dstep*blktrace(blk,dX,dZ); 
       gapcorr = (normb*normC)*trXZcorr;
       mucorr  = trXZcorr/n;
       time(6)=cputime;
       step_corrtime = step_corrtime + (time(6)-time(5)); 
%%
%%-----------------------------------------
%%  stopping criteria for corrector step
%%-----------------------------------------
%%
       if (pstep < steptol) | (dstep < steptol); 
          fprintf('\n Stop: steps in corrector too short,');
          fprintf('  pstep = %3.2e,  dstep = %3.2e\n',pstep,dstep);
          CPUtime(iter) = cputime-tstart; 
          termcode = -2; 
          break;
       end;
       %% test for slow progress when mu is small
       %%
       if (iter >= 2); 
          idx = [max(2,iter-5) : iter];
          corr_convg_rate = mean(gaphist(idx)./gaphist(idx-1));
          idx = [max(2,iter-2) : iter];
          corr_slow = all(gaphist(idx)./gaphist(idx-1) > 0.5); 
       end; 
       if (mu < 1e-6 & infeas_meas < 1e-6); 
          if (mucorr/mu < min(1,5*corr_convg_rate)) & (~corr_slow); 
             if (mucorr < 3*mupred);
                X = ops(X,'+',dX,pstep);  y = y+dstep*dy;  
                Z = ops(Z,'+',dZ,dstep); 
                gap = gapcorr; mu = mucorr;  
             else;
                fprintf('  update to good predictor point\n'); 
                X = Xpred;  y = ypred;  Z = Zpred; 
                gap = gappred; mu = mupred;
             end;
          else;
             fprintf('\n  lack of progress in corrector:');
             fprintf(' mucorr/mu = %3.2f, corr_convg_rate = %3.2f\n',...
                       mucorr/mu,corr_convg_rate); 
             if (mupred/mu < 0.2); 
                fprintf('  update to good predictor point\n'); 
                X = Xpred;  y = ypred;  Z = Zpred; 
                gap = gappred; mu = mupred;
             else; 
                fprintf('  Stop: lack of progress in corrector.\n');
                CPUtime(iter) = cputime-tstart; 
                termcode = -1; 
                break;
             end; 
          end; 
       else;
          X = ops(X,'+',dX,pstep);  y = y+dstep*dy;  
          Z = ops(Z,'+',dZ,dstep);
          gap = gapcorr;  mu = mucorr; 
       end; 
       clear Xpred  ypred  Zpred; 
    end; %% end of corrector step. 
%%
%%---------------------------------------------------------------
%% compute rp, Rd, infeasibities, etc.
%%---------------------------------------------------------------
%%
    clear invXchol invZchol  dX  dy  dZ; 
    trXZ = blktrace(blk,X,Z); 
    gap = (normb*normC)*trXZ;
    mu  = trXZ/n;
    AX  = blktrace(blk,A,X);
    rp  = b-AX;
    ZpATy = ops(Z,'+',Asum(blk,A,y,iscmpA));
    ZpATynorm = ops(ZpATy,'norm');
    Rd  = ops(C,'-',ZpATy); 
    clear ZpATy;
    obj = (normb*normC)*[blktrace(blk,C,X)  b'*y]; 
    rel_gap = gap/max(1,mean(abs(obj))); 
    pri_infeas  = norm(rp)/normb2; 
    dual_infeas = ops(Rd,'norm')/normC2;
  %%pri_infeas  = normb*norm(normA.*rp)/normb0; %% for original data
  %%dual_infeas = normC*ops(Rd,'norm')/normC0;  %% for original data
    infeas_meas = max(pri_infeas,dual_infeas); 
    gaphist(iter+1)     = gap;  
    infeashist(iter+1)  = infeas_meas;
    stephist(iter+1) = min(pstep,dstep); 
    CPUtime(iter) = cputime-tstart;
    time(7) = cputime;
    if predcorr; 
       misctime_iter = time(7)-time(6);
       misctime = misctime + misctime_iter; 
    else;
       misctime_iter = time(7)-time(4); 
       misctime = misctime + misctime_iter; 
    end; 
%%
    if printyes;
      fprintf('%2.0f  %4.3f %4.3f',iter,pstep,dstep);
      fprintf(' %2.1e %2.1e  %2.1e',pri_infeas,dual_infeas,gap);
      fprintf(' %- 7.6e  %3.3f\n',mean(obj),sigma);
    end;
%%
%%--------------------------------------------------
%% check convergence.
%%--------------------------------------------------
%%
     if (max(rel_gap,infeas_meas) < gaptol);
        fprintf('\n Stop: max(relative gap, infeasibilities) < %3.2e',gaptol);
        termcode = 0;
        break;
     end;
     min_pri_infeas = min(infeashist(1:iter)); 
     pri_infeas_bad = pri_infeas_bad + ...
        (pri_infeas > max(1e-10,min_pri_infeas) & (min_pri_infeas < 1e-2));
     if (infeas_meas < 1e-6 | pri_infeas_bad) & (mu < 1e-3); 
        if (mu < 1e-8);
           idx = [max(1,iter-1): iter];
        elseif (mu < 1e-4);
           idx = [max(1,iter-2): iter]; 
        else
           idx = [max(1,iter-3): iter];
        end;
        gap_ratio = gaphist(idx+1)./gaphist(idx); 
        idx2 = [max(1,iter-4): iter]; 
        gap_slowrate = max(0.5, 1.2*mean(gaphist(idx2+1)./gaphist(idx2)));
        gap_slow = ( all(gap_ratio > gap_slowrate) | ...
                     (pri_infeas_bad & any(gap_ratio > 0.99)) ) ...
                   & (gap/maxblk < 1e-3 | rel_gap < 1e-3);
        if (max(infeashist) > 1e-6) & (mu < 1e-4); 
           step_short = all(stephist([iter:iter+1]) < 0.1) | ...
                        (stephist(iter+1) < 0.05); 
        else; 
           %% if starting point is nearly feasible, do not test for 
           %% short step here. 
           step_short = 0; 
        end;
        if (rel_gap < 5*infeas_meas); 
           fprintf('\n Stop: relative gap < 5*infeasibility');  
           termcode = 0;
           break;  
        end;
        if (gap_slow) | (step_short);
           fprintf('\n Stop: progress is too slow. ');
           termcode = -5; 
           break;
        end;  
     end;
     if (obj(2)/(normb*normC) > ZpATynorm / max(inftol,1e-13));
        fprintf('\n Stop: primal problem is suspected of being infeasible');
        termcode = 1;
        break;
     end;
     if (-obj(1)/(normb*normC) > norm(AX) / max(inftol,1e-13));
        fprintf('\n Stop: dual problem is suspected of being infeasible');
        termcode = 2;
        break;
     end;
  end;
%%---------------------------------------------------------------
%% end of main loop
%%---------------------------------------------------------------
%%
   if (termcode == -6);
      fprintf('\n Stop: maximum number of iterations reached \n');
   end;
%%
%%---------------------------------------------------------------
%% unscale and produce infeasibility certificates if appropriate
%%---------------------------------------------------------------
%%
   if (termcode <= 0),
      if (scale_data == 1),
         X = ops(normb,'*',X); y = normC*y./normA; Z = ops(normC,'*',Z);
      end;
      if (nargout >= 8); Xiter = X; yiter = y; Ziter = Z; end; 
   end;
   if (termcode == 1);
      if (scale_data == 1),
         b = normb*b; b = b.*normA; y = y./normA;
         X = ops(normb,'*',X);
         if (nargout >= 8); 
            yiter = normC*y; Ziter = ops(normC,'*',Z); Xiter = X; 
         end; 
      else;
         if (nargout >= 8); Xiter = X; yiter = y; Ziter = Z; end;
      end;
      rby = 1/(b'*y); y = rby*y; Z = ops(Z,'*',rby);
      norm_resid = ZpATynorm * rby;
   end;  
   if (termcode == 2);
      if (scale_data == 1),
         C = ops(C,'*',normC);
         y = normC*y./normA; Z = ops(normC,'*',Z);
         if (nargout >= 8); 
            Xiter = ops(normb,'*',X); yiter = y; Ziter = Z;
         end; 
      else;
         if (nargout >= 8);  Xiter = X; yiter = y; Ziter = Z; end; 
      end;
      tCX = blktrace(blk,C,X);
      X = ops(X,'*',1/(-tCX));
      norm_resid = norm(normA.*AX)/(-tCX);
   end;
%%
%%-----------------------------------------
%%  print summary
%%-----------------------------------------
%%
   fprintf('\n----------------------------------------------------\n');
   fprintf(' number of iterations   = %2.0f\n',iter);
   if (termcode <= 0),
      info = [termcode iter gap infeas_meas sum(CPUtime) ];
      fprintf(' gap                    = %3.2e\n',gap);
      fprintf(' relative gap           = %3.2e\n',rel_gap);
      fprintf(' infeasibilities        = %3.2e\n',infeas_meas);
   elseif (termcode == 1),
      info = [termcode iter gap norm_resid sum(CPUtime) ];
      fprintf(' residual of primal infeasibility      \n')
      fprintf(' certificate (y,Z)      = %3.2e\n',norm_resid);
   elseif (termcode == 2),
      info = [termcode iter gap norm_resid sum(CPUtime) ];
      fprintf(' residual of dual infeasibility        \n')
      fprintf(' certificate X          = %3.2e\n',norm_resid);
   end;
   fprintf(' Total CPU time         = %3.1f\n',sum(CPUtime));
   fprintf(' CPU time per iteration = %3.1f\n',sum(CPUtime)/iter);
   fprintf(' termination code       = %2.0f\n',termcode);
   fprintf('----------------------------------------------------\n');
   fprintf('  Percentage of CPU time spent in various parts \n'); 
   fprintf('----------------------------------------------------\n');
   fprintf('  chol   pred   steplen   corr   steplen   misc\n')
   tt = [choltime predtime step_predtime corrtime step_corrtime misctime];
   tt = tt/sum(tt)*100; 
   fprintf('  %3.1f    %3.1f   %3.1f',tt(1),tt(2),tt(3));
   fprintf('       %3.1f   %3.1f       %3.1f\n',tt(4),tt(5),tt(6)); 
   fprintf('----------------------------------------------------\n');
%%=============================================================================
