%%*****************************************************************************
%% sdphlf: solve an SDP by homogeneous path-following method. 
%%
%%  vers = 1:  AHOhlf-*-Sch. 
%%  vers = 2:  HKMhlf-*-Sch.
%%  vers = 3:  NThlf-*-Sch.
%%  vers = 4:  GThlf-*-Sch.
%%
%%  functions:  AHOpredhlf, AHOcorrhlf    
%%              HKMpredhlf, HKMcorrhlf
%%              NTpredhlf,  NTcorrhlf
%%              validate   [validate SDP data]  
%%              scaling    [scale SDP data]    
%%              blkchol    [Cholesky factorizations]
%%              corrprim   [correct primal infeasibility]  
%%              steplength [calculate maximum feasible steplength]
%%
%%  [obj,X,y,Z,gaphist,infeashist,info,Xiter,yiter,Ziter,tau,kap] = ...
%%              sdphlf(blk,A,C,b,X0,y0,Z0,tau0,kap0,OPTIONS);
%%
%%  Input:  blk: a cell array describing the block diagonal structure
%%               of SDP data.
%%            A: a cell array with A{k} = A_k, k = 1:m, 
%%          C,b: data for the SDP instance.
%%   (X0,y0,Z0): an initial iterate.
%%  (tau0,kap0): an initial iterate. Typically, one chooses tau0, kap0 = 1
%%               (if they are not given, this is the default).
%%      OPTIONS: a structure specifying parameters required in sdp.m
%%               (if it is not given, the default in parameters.m is used). 
%%
%%  Output: obj  = [blktrace(blk,C,X) b'*y]/tau; 
%%          (X,y,Z): an approximately optimal solution, or
%%                   a certificate of primal or dual infeasibility.
%%          gaphist = history of reduction in <X,Z>.
%%          infeashist = history of reduction in the infeasibility measure. 
%%          info = [termination-code  iters  gap  infeasibity  CPUtime].
%%          Xiter, yiter, Ziter: values of X, y, and Z at the final iterate.
%%          tau, kap: values of tau and kappa at the final iterate.
%%
%%----------------------------------------------------------------------------
%%  The OPTIONS structure specifies the required parameters: 
%%      vers  gam  predcorr  expon  gaptol  inftol  steptol  maxit 
%%      sw2PC_tol  use_corrprim     printyes
%%      (all have default values set in parameters.m).
%%
%%  vers = version of the direction to use. 
%%  gam  = steplength parameter
%%         [default: gam = 0, it is adaptively selected within the algorithm]. 
%%  predcorr = 1, if use Mehrotra-type predictor-corrector, else = 0
%%         [default 1]. 
%%  expon = exponent in decrease of centering parameter sigma
%%         [default: [3 1 1 2], for AHO, HKM, NT, and GT respectively]. 
%%  gaptol = tolerance for duality gap as a fraction of the value of the 
%%         objective functions    
%%         [default 1e-8]
%%  inftol = tolerance for stopping due to suspicion of infeasibility
%%         [default 1e-13]
%%  steptol = tolerance for stopping due to small steps 
%%         [default 1e-6].
%%  maxit = maximum number of iteration allowed
%%         [default 50].
%%  sw2PC_tol = use PC rule only if infeasibility < sw2PC_tol 
%%         [default inf].
%%  use_corrprim = 1, if correct primal infeasibility each step 
%%         (after feasibility attained), else = 0 
%%         [default 0].
%%  printyes = 1, if want to display result in each iteration, else = 0
%%         [default 1].
%%  scale_data = 1, if want to scale the data before solving the problem, 
%%         else 0
%%         [default 0].
%%  schurfun = [], if no user supplied routine to compute the Schur matrix, 
%%             or, a string containing the file name where the Schur matrix
%%             is computed,
%%             e.g.  schurfun = 'mcpHKMsch(X,Zinv)', or 
%%                   schurfun = 'thetaHKMsch(X,Zinv,Rd,schurfun_parms)'.  
%%  schurfun_parms = [], if no user supplied routine to compute the Schur 
%%             matrix, or, a cell array containing the parameters need in 
%%             the user supplied Schur routine. 
%%  randnstate = state to be used for matlab 5 random normal generator
%%         [default 0].
%%*****************************************************************************
%% SDPT3: version 2.1
%% Copyright (c) 1997 by
%% K.C. Toh, M.J. Todd, R.H. Tutuncu
%% Last modified: 7 Jul 99
%%****************************************************************************

   function [obj,X,y,Z,gaphist,infeashist,info,Xiter,yiter,Ziter,tau,kap] ...
            = sdphlf(blk,A,C,b,X0,y0,Z0,tau0,kap0,OPTIONS);
%%
%%-----------------------------------------
%% get parameters from the OPTIONS structure. 
%%-----------------------------------------
%%
   if (nargin <=9);  parameters; end; 
   if (nargin <=8);  kap0 = 1; end;
   if (nargin <=7);  tau0 = 1; end;
   if isstruct(tau0); 
      error('tau0 is set to the structure OPTIONS');  
   end;
   warning off; 

   vers           = OPTIONS.vers; 
   gam            = OPTIONS.gam; 
   predcorr       = OPTIONS.predcorr; 
   expon          = OPTIONS.expon;
   gaptol         = OPTIONS.gaptol;
   inftol         = OPTIONS.inftol;
   steptol        = OPTIONS.steptol;
   maxit          = OPTIONS.maxit;
   sw2PC_tol      = OPTIONS.sw2PC_tol;
   use_corrprim   = OPTIONS.use_corrprim;
   printyes       = OPTIONS.printyes;
   scale_data     = OPTIONS.scale_data;
   schurfun       = OPTIONS.schurfun;
   schurfun_parms = OPTIONS.schurfun_parms;
   randnstate      = OPTIONS.randnstate; 
%%
   if isempty(gam) | isempty(predcorr) | isempty(use_corrprim) | ...
      ~all([expon sw2PC_tol gaptol steptol]);
      fprintf('\n The variable OPTIONS in ** parameters.m **');
      fprintf(' require default settings. \n'); 
      termcode = -10;
      break; 
   end;
   if all(vers-[1 2 3 4]); error('*** vers must be 1,2,3 or 4***'); end; 
   pri_infeas_tol = 1e-12; 
%%
   if use_corrprim ~= 0; 
      fprintf('\n *** not advisable to correct primal infeasibility in\n') 
      fprintf(' *** homogeneous algorithms: setting use_corrprim = 0\n'); 
      use_corrprim = 0; 
   end;
%%
%%-----------------------------------------
%% validate data
%%-----------------------------------------
%%
   X = X0; y = y0; Z = Z0; kap = kap0; tau = tau0; 
   clear X0 y0 Z0; 
   if ~iscell(C); tmp = C; clear C; C{1} = tmp; end;
   if ~iscell(X); tmp = X; clear X; X{1} = tmp; end;
   if ~iscell(Z); tmp = Z; clear Z; Z{1} = tmp; end;
   clear tmp; 

   spdensity = 0.3;
   [A,C,X,Z,isspA,iscmpA,nnzA] = validate(blk,A,C,b,X,y,Z,spdensity);
%%
%%-----------------------------------------
%% find the combined list of non-zero
%% elements of Aj, j = 1:k, for each k. 
%%-----------------------------------------
%%
   [listA,splistA,permA,listRd] = nonzerolist(blk,A,spdensity,C);
%%
%%
%%-----------------------------------------
%% compute A*A' to determine if the matrices
%% Ak are linearly independent. If they are
%% dependent, preprocess to get a linearly
%% independent set.  
%%-----------------------------------------
%%
   m = length(b);  AAt = zeros(m); 
   for k = [1:m];  
       tmp = blktrace(blk,A(:,1:k),A(:,k));  
       AAt(1:k,k) = tmp;  AAt(k,1:k) = tmp';  
   end;
   if (nnz(AAt) <= spdensity*m*m); AAt = sparse(AAt); end;
   permUA = symrcm(AAt); 
   [UA,indef] = chol(AAt(permUA,permUA));
   if indef | (min(abs(diag(UA))) < 1e-6); 
      fprintf('\n matrices Ak are linearly dependent.'); 
      [A,b,y,indeprows,feasible] = preprocess(blk,A,b,y,isspA);
      if use_corrprim;
         AAt = AAt(indeprows,indeprows); 
         [UA,indef] = chol(AAt(permUA,permUA));
      end; 
   end; 
   if (~use_corrprim); clear UA  permUA; UA = []; permUA = []; end; 
   clear AAt; 
%%
%%-----------------------------------------
%% scale SDP data 
%%-----------------------------------------
%%
   normb0 = max(1,norm(b)); normC0 = max(1,ops(C,'norm')); 
   if (scale_data == 1),
      [A,C,b,X,y,Z,normA,normC,normb] = scaling(blk,A,C,b,X,y,Z);
      if ~isempty(UA); 
         UA = UA./(ones(size(normA))*normA'); 
      end;
   else;
      normA = ones(m,1); normC = 1; normb = 1;
   end;
%%
%%-----------------------------------------
%% decomposition of A_k 
%%-----------------------------------------
%% 
   if (vers == 1 | vers == 4); 
      for p = 1:size(blk,1) 
         if (isspA(p) == 0); [LA(p,:),TA(p,:),PA(p,:)] = aasen(A(p,:)); end;
      end; 
   elseif (vers == 3); 
      for p = 1:size(blk,1) 
         if (isspA(p) == 0); AR(p,:) = Atriu(A(p,:)); end;
      end; 
   end;
%%
%%-----------------------------------------
%% initialization
%%-----------------------------------------
%%
   m = length(b);  n = ops(C,'getM');
   rp  = [tau*b-blktrace(blk,A,X) ;  blktrace(blk,C,X)-b'*y+kap];   
   Rd  = ops(ops(tau,'*',C),'-',ops(Z,'+',Asum(blk,A,y,iscmpA))); 
   obj = (normb*normC)*[blktrace(blk,C,X) b'*y]/tau;
   trXZ = blktrace(blk,X,Z); 
   gap = (normb*normC)*trXZ/tau^2; 
   mu  = (blktrace(blk,X,Z) + kap*tau)/(n+1);  mu0 = max(1,mu);
   rel_gap = gap/max(1,mean(abs(obj)));
   normb2 = max(1,norm(b)); 
   normC2 = max(1,ops(C,'norm')); 
   pri_infeas  = norm(rp(1:m)/tau)/normb2; 
   dual_infeas = ops(ops(Rd,'/',tau),'norm')/normC2; 
 %%pri_infeas  = normb*norm(normA.*rp(1:m)/tau)/normb0;    %% for original data
 %%dual_infeas = normC*ops(ops(Rd,'/',tau),'norm')/normC0; %% for original data
   infeas_meas = max(pri_infeas,dual_infeas);
   pstep = 0; dstep = 0;  step_pred = 0; 
   pred_covg = 1; corr_convg_rate = 1; 
   termcode = -6; 
   gaphist(1)  = gap; 
   infeashist(1) = infeas_meas;
   pri_infeas_bad = 0; 
   predcorr0 = predcorr;
%%
%%-----------------------------------------
%% display parameters, and initial info
%%-----------------------------------------
%% 
   fprintf('\n*************************************************');
   fprintf('******************\n');
   fprintf('         Homogeneous self-dual algorithms'); 
   if ~isempty(schurfun);  
      fprintf('\n\nuser supplied Schur routine: '); 
      fprintf(schurfun); 
   end;
   fprintf('\n*************************************************');
   fprintf('******************\n');
   fprintf(' version  predcorr  gam  expon  use_corrprim  sw2PC_tol  scale_data\n');
   fprintf('   %1.0f        %1.0f      %4.3f',vers,predcorr,gam);
   fprintf('    %1.0f        %1.0f         %2.1e          %1.0f\n',...
           expon(vers),use_corrprim,sw2PC_tol,scale_data);
   if printyes;
      fprintf('\nit pstep dstep p_infeas d_infeas  gap')
      fprintf('         obj      sigma\n');
      fprintf('------------------------------------------------');
      fprintf('-------------------\n');
      fprintf('%2.0f  %4.3f %4.3f',0,0,0);
      fprintf(' %2.1e %2.1e %2.1e',pri_infeas,dual_infeas,gap);
      fprintf(' %- 7.6e\n',mean(obj));
   end;
%%
%%---------------------------------------------------------------
%% start main loop
%%---------------------------------------------------------------
%%
   choltime = 0; predtime = 0; step_predtime = 0;
   corrtime = 0; step_corrtime = 0; misctime = 0; 
   for iter = 1: maxit;  
       flops(0); 
       time = zeros(1,7); 
       tstart = cputime;
       time(1) = cputime;
       [Xchol,indef(1)] = blkchol(blk,X); 
       [Zchol,indef(2)] = blkchol(blk,Z); 
       invXchol = cell(size(X));  %% must initialize.
       invZchol = cell(size(Z));       
       time(2) = cputime;
       choltime_iter = time(2)-time(1); 
       choltime = choltime + choltime_iter;
       if any(indef);  
          fprintf('\n Stop: X, Z are not both positive definite');
          termcode = -3;
          break; 
       end;
%%
%%---------------------------------------------------------------
%% predictor step.
%%---------------------------------------------------------------
%%
     if predcorr & (infeas_meas < sw2PC_tol); 
        sigma = 0; 
     else; 
        sigma = 1-0.9*min(pstep,dstep); 
        if (iter == 1); sigma = 0.5; end; 
     end; 
     eta = 1-sigma; 
%%
%%-----------------------------------------
%% determine which way to compute the Schur 
%% matrix from the CPU time taken via two
%% different ways in the 3nd and 4th iteration
%% when the data correspond to the pth block 
%% is dense. 
%%-----------------------------------------
%% 
     if isempty(schurfun); 
        schur_opt = 2*ones(1,size(blk,1)); 
        dense_idx = find(isspA == 0); 
        if ~isempty(dense_idx); 
           if (iter > 1); schur_cpu_hist(iter-1,:) = schur_cpu; end;
           if ~all(iter-[1 3]); 
              schur_opt(dense_idx) = ones(1,length(dense_idx));  
           elseif ~all(iter-[2 4]); 
              schur_opt(dense_idx) = 2*ones(1,length(dense_idx)); 
           else; 
             tmp=schur_cpu_hist(3,dense_idx)./(schur_cpu_hist(4,dense_idx)+.1);
             schur_opt(dense_idx) = (tmp > 1) + 1; 
           end;
        end;
     else;
        schur_opt = schurfun; 
     end;
%%
     if (vers == 1);      
       [dX,dy,dZ,dtau,dkap,hRd,V,d,B,hX,Lsch,Usch,Psch,schur_cpu] = ...
          AHOpredhlf(blk,A,C,b,rp,Rd,sigma,mu,eta,schur_opt,schurfun_parms,...
          iscmpA,normA,listA,permA,listRd,tau,kap,X,Z,LA,TA,PA);
        invXchol = ops(Xchol,'inv');  clear Xchol;
        invZchol = ops(Zchol,'inv');  clear Zchol; 
     elseif (vers == 2);
        invZchol = ops(Zchol,'inv');  clear Zchol; 
        [dX,dy,dZ,dtau,dkap,hRd,Zinv,Uchol,Lsch,Usch,Psch,schur_cpu] = ...
           HKMpredhlf(blk,A,C,b,rp,Rd,sigma,mu,eta,schur_opt,schurfun_parms,...
           iscmpA,normA,listA,permA,listRd,splistA,tau,kap,X,Xchol,invZchol);
        invXchol = ops(Xchol,'inv');  clear Xchol;
     elseif (vers == 3);
        invZchol = ops(Zchol,'inv'); 
        [dX,dy,dZ,dtau,dkap,hRd,W,G,sv,Uchol,Lsch,Usch,Psch,schur_cpu] = ...
          NTpredhlf(blk,A,C,b,rp,Rd,sigma,mu,eta,schur_opt,schurfun_parms,...
          iscmpA,normA,listA,permA,listRd,splistA,tau,kap,X,Zchol,invZchol,AR);
        clear Zchol; 
        invXchol = ops(Xchol,'inv');  clear Xchol;
     elseif (vers == 4); 
        invZchol = ops(Zchol,'inv'); 
        [dX,dy,dZ,dtau,dkap,hRd,ZH,invZH,d,sv2,phi2,psi2,Uchol,Lsch,Usch,...
           Psch,schur_cpu] = GTpredhlf(blk,A,C,b,rp,Rd,sigma,mu,eta,...
           schur_opt,schurfun_parms,iscmpA,normA,listA,permA,listRd,...
           tau,kap,X,Zchol,invZchol,LA,TA,PA);
        clear Zchol; 
        invXchol = ops(Xchol,'inv');  clear Xchol;
     end; 
     if any([dy; dtau] == inf);
        disp('*** Schur complement system is singular ***');  
        CPUtime(iter) = cputime-tstart; 
        termcode = -4;
        break;
     end; 
     time(3) = cputime;
     predtime_iter = time(3)-time(2); 
     predtime = predtime + predtime_iter;
%%
%%-----------------------------------------
%% step-lengths for predictor step
%%-----------------------------------------
%%
     if (gam == 0); 
        gamused = 0.9 + 0.09*min(pstep,dstep); 
     else
        gamused = gam;
     end; 
     if use_corrprim & (pri_infeas < pri_infeas_tol)
        dX = corrprim(blk,A,dX,UA,permUA); end; 
     Xstep = steplength(blk,invXchol,dX,randnstate); 
     Zstep = steplength(blk,invZchol,dZ,randnstate);
     kapstep = max( (dkap<0)*(-kap/(dkap-eps)), (dkap>=0)*1e6 );
     taustep = max( (dtau<0)*(-tau/(dtau-eps)), (dtau>=0)*1e6 ); 
     pstep = min(1,gamused*min([Xstep,kapstep,taustep]));
     dstep = min(1,gamused*min([Zstep,kapstep,taustep]));
     minstep = min(pstep,dstep); 
     nu = blktrace(blk,Z,dX)*(pstep>dstep) + blktrace(blk,X,dZ)*(pstep<dstep) ...
         + kap*dtau*(dtau>0) + tau*dkap*(dtau<0);           
     muscaled = mu*(1-minstep*(1-sigma)) + abs(pstep-dstep)*nu/(n+1); 
     muscaled = 1/((tau+pstep*dtau)*(tau+dstep*dtau))*muscaled;       
     if (dtau < -eta*tau) | (muscaled > mu/tau^2); 
        pstep = minstep; dstep = minstep;         
     end;
     step_pred = min(pstep,dstep);  
     time(4) = cputime;
     step_predtime = step_predtime + time(4)-time(3); 
%%
%%-----------------------------------------
%% updates for predictor step 
%%-----------------------------------------
%%
     [taupred,idx] = min([tau+pstep*dtau,tau+dstep*dtau]); 
     if (idx == 1); kappred = kap + pstep*dkap; 
     else;          kappred = kap + dstep*dkap; end; 
     Xpred = ops(taupred/(tau+pstep*dtau),'*',ops(X,'+',dX,pstep));  
     ypred = (taupred/(tau+dstep*dtau))*(y+dstep*dy);  
     Zpred = ops(taupred/(tau+dstep*dtau),'*',ops(Z,'+',dZ,dstep)); 
     trXZpred = taupred^2/(tau+pstep*dtau)/(tau+dstep*dtau)* ...
                (trXZ + pstep*blktrace(blk,dX,Z) + dstep*blktrace(blk,X,dZ)...
                 + pstep*dstep*blktrace(blk,dX,dZ)); 
     gappred = (normb*normC)*trXZpred/(taupred^2); 
     mupred  = (trXZpred+kappred*taupred)/(n+1);
     mupredhist(iter) = mupred; 
%%
%%-----------------------------------------
%%  stopping criteria for predictor step.
%%-----------------------------------------
%%
     if (pstep < steptol) | (dstep < steptol);
        fprintf('\n Stop: steps in predictor too short:');
        fprintf(' pstep = %3.2e,  dstep = %3.2e\n',pstep,dstep);
        CPUtime(iter) = cputime-tstart; 
        termcode = -2; 
        break; 
     end;
     %% test for slow progress when mu is small
     %%
     if (iter >= 2); 
        idx = [max(2,iter-5) : iter];
        pred_convg_rate = mean(mupredhist(idx)./mupredhist(idx-1));
        idx = [max(2,iter-2) : iter];
        pred_slow = all(mupredhist(idx)./mupredhist(idx-1) > 0.2); 
     end; 
     if (~predcorr) | (predcorr & (infeas_meas >= sw2PC_tol)); 
        if (mu < 1e-8 & infeas_meas < 1e-8) ...
           & ((mupred/mu > 5*pred_convg_rate) | (pred_slow));
           fprintf('\n Stop: lack of progress in predictor:');
           fprintf(' mupred/mu = %3.2f, pred_convg_rate = %3.2f\n',...
                     mupred/mu,pred_convg_rate);
           CPUtime(iter) = cputime-tstart; 
           termcode = -1; 
           break;           
        else;
           X = ops(taupred/(tau+pstep*dtau),'*',ops(X,'+',dX,pstep));  
           y = (taupred/(tau+dstep*dtau))*(y+dstep*dy);  
           Z = ops(taupred/(tau+dstep*dtau),'*',ops(Z,'+',dZ,dstep)); 
           tau = taupred;  kap = kappred; 
           gap = gappred;  mu = mupred; 
        end;
     end;
%%
%%---------------------------------------------------------------
%%  corrector step.
%%---------------------------------------------------------------
%%
    if predcorr & (infeas_meas < sw2PC_tol);

       Xpred = ops(taupred/(tau+pstep*dtau),'*',ops(X,'+',dX,pstep)); 
       ypred = ops(taupred/(tau+dstep*dtau),'*',ops(y,'+',dy,dstep));  
       Zpred = ops(taupred/(tau+dstep*dtau),'*',ops(Z,'+',dZ,dstep)); 
       mupred = (blktrace(blk,Xpred,Zpred) + kappred*taupred)/(n+1);
%%
       if (mu > 1e-6); 
          if (vers == 1) & (step_pred < 1/sqrt(3)); 
             expon_used = 1; 
          else
             expon_used = max(expon(vers),3*step_pred^2); 
          end; 
       else
          expon_used = max(1,min(expon(vers),3*step_pred^2)); 
       end;
       sigma = min( 1, (mupred/mu)^expon_used ); 
       eta = 1-sigma; 
%%      
       if (vers == 1); 
          [dX,dy,dZ,dtau,dkap] = AHOcorrhlf(blk,dX,dZ,dtau,dkap,A,C,b,rp,Rd,...
             hRd,sigma,mu,eta,Lsch,Usch,Psch,iscmpA,tau,kap,schur_opt,X,Z,V,...
             d,B,hX);
          clear Lsch Usch Psch V B hX; 
       elseif (vers == 2); 
          [dX,dy,dZ,dtau,dkap] = HKMcorrhlf(blk,dX,dZ,dtau,dkap,A,C,b,rp,Rd,...
              hRd,sigma,mu,eta,Uchol,Lsch,Usch,Psch,iscmpA,tau,kap,X,Zinv);
          clear Uchol Lsch Usch Psch Zinv; 
       elseif (vers == 3);
           [dX,dy,dZ,dtau,dkap] = NTcorrhlf(blk,dX,dZ,dtau,dkap,A,C,b,rp,Rd,...
               hRd,sigma,mu,eta,Uchol,Lsch,Usch,Psch,iscmpA,tau,kap,W,G,sv);
           clear Uchol Lsch Usch Psch W G; 
       elseif (vers == 4);
           [dX,dy,dZ,dtau,dkap] = GTcorrhlf(blk,dX,dZ,dtau,dkap,A,C,b,rp,Rd,...
               hRd,sigma,mu,eta,Uchol,Lsch,Usch,Psch,iscmpA,tau,kap,ZH,...
               invZH,d,sv2,phi2,psi2);
           clear Uchol Lsch Usch Psch ZH invZH; 
       end;
       if any([dy; dtau] == inf);
          disp('*** Schur complement system is singular ***');  
          CPUtime(iter) = cputime-tstart; 
          termcode = -4;
          break;
       end; 
       step_corr = min(pstep,dstep); 
       time(5) = cputime;
       corrtime_iter = time(5)-time(4);
       corrtime = corrtime + corrtime_iter;
%%
%%-----------------------------------
%% step-lengths for corrector step
%%-----------------------------------
%%
       if (gam == 0); 
          gamused = 0.9 + 0.09*min(pstep,dstep); 
       else
          gamused = gam;
       end;  
       if use_corrprim & (pri_infeas < pri_infeas_tol)
          dX = corrprim(blk,A,dX,UA,permUA); end; 
       Xstep = steplength(blk,invXchol,dX,randnstate); 
       Zstep = steplength(blk,invZchol,dZ,randnstate); 
       kapstep = max( (dkap<0)*(-kap/(dkap-eps)), (dkap>=0)*1e6);
       taustep = max( (dtau<0)*(-tau/(dtau-eps)), (dtau>=0)*1e6); 
       pstep = min(1,gamused*min([Xstep,kapstep,taustep]));
       dstep = min(1,gamused*min([Zstep,kapstep,taustep]));
       minstep = min(pstep,dstep); 
       nu = blktrace(blk,Z,dX)*(pstep>dstep) + blktrace(blk,X,dZ)*(pstep<dstep) ...
            + kap*dtau*(dtau>0) + tau*dkap*(dtau<0);     
       muscaled = mu*(1-minstep*(1-sigma)) + abs(pstep-dstep)*nu/(n+1); 
       muscaled = 1/((tau+pstep*dtau)*(tau+dstep*dtau))*muscaled;       
       if (dtau < -eta*tau) | (muscaled > mu/tau^2); 
          pstep = minstep; dstep = minstep; 
       end;
%%
%%-----------------------------------
%% updates for corrector step
%%-----------------------------------
%%
       [taucorr,idx] = min([tau+pstep*dtau,tau+dstep*dtau]); 
       if (idx == 1); kapcorr = kap + pstep*dkap; 
       else;          kapcorr = kap + dstep*dkap; end; 
       trXZcorr = taucorr^2/(tau+pstep*dtau)/(tau+dstep*dtau) * ...
                (trXZ + pstep*blktrace(blk,dX,Z) + dstep*blktrace(blk,X,dZ)...
                 + pstep*dstep*blktrace(blk,dX,dZ));     
       gapcorr = (normb*normC)*trXZcorr/(taucorr^2); 
       mucorr  = (trXZcorr+kapcorr*taucorr)/(n+1);
       mucorrhist(iter) = mucorr; 
       time(6)=cputime;
       step_corrtime = step_corrtime + time(6)-time(5); 
%%
%%-----------------------------------------
%%  stopping criteria for corrector step
%%-----------------------------------------
%%
       if (pstep < steptol) | (dstep < steptol); 
          fprintf('\n Stop: steps in corrector too short,');
          fprintf('  pstep = %3.2e,  dstep = %3.2e\n',pstep,dstep);
          CPUtime(iter) = cputime-tstart; 
          termcode = -2; 
          break;
       end;
       %% test for slow progress when mu is small
       %%
       if (iter >= 2); 
          idx = [max(2,iter-5) : iter];
          corr_convg_rate = mean(gaphist(idx)./gaphist(idx-1));
          idx = [max(2,iter-2) : iter];
          corr_slow = all(gaphist(idx)./gaphist(idx-1) > 0.5); 
       end; 
       if (mu < 1e-6) & (infeas_meas < 1e-6); 
          if (mucorr/mu < min(1.2,5*corr_convg_rate)) & (~corr_slow); 
             if (mucorr < 3*mupred);
                X = ops(taucorr/(tau+pstep*dtau),'*',ops(X,'+',dX,pstep));  
                y = (taucorr/(tau+dstep*dtau))*(y+dstep*dy);  
                Z = ops(taucorr/(tau+dstep*dtau),'*',ops(Z,'+',dZ,dstep));  
                tau = taucorr; kap = kapcorr;
                gap = gapcorr; mu = mucorr; 
             else; 
                fprintf('    update to good predictor point\n'); 
                X = Xpred; y = ypred; Z = Zpred; 
                tau = taupred; kap = kappred;
                gap = gappred; mu = mupred;
             end;
          else; 
             fprintf('\n   lack of progress in corrector:');
             fprintf(' mucorr/mu = %3.2f, corr_convg_rate = %3.2f\n',...
                       mucorr/mu,corr_convg_rate); 
             if (mupred/mu < 0.2); 
                fprintf('    update to good predictor point\n'); 
                X = Xpred; y = ypred; Z = Zpred; 
                tau = taupred; kap = kappred;
                gap = gappred; mu = mupred;
             else; 
                fprintf('   Stop: lack of progress in corrector.\n');
                CPUtime(iter) = cputime-tstart; 
                termcode = -1; 
                break;
             end; 
          end; 
       else;            
          X = ops(taucorr/(tau+pstep*dtau),'*',ops(X,'+',dX,pstep));  
          y = (taucorr/(tau+dstep*dtau))*(y+dstep*dy);  
          Z = ops(taucorr/(tau+dstep*dtau),'*',ops(Z,'+',dZ,dstep));  
          tau = taucorr; kap = kapcorr;
          gap = gapcorr; mu = mucorr; 
       end; 
       clear Xpred ypred Zpred; 
    end; %% end of corrector step.
%%
%%---------------------------------------------------------------
%% compute infeasibities, etc.
%%---------------------------------------------------------------
%%
    clear invXchol  invZchol  dX  dy  dZ; 
    trXZ = blktrace(blk,X,Z);
    gap = (normb*normC)*trXZ/(tau^2); 
    mu  = (trXZ+kap*tau)/(n+1);
    AX  = blktrace(blk,A,X);
    rp  = [tau*b-AX ;  blktrace(blk,C,X)-b'*y+kap]; 
    ZpATy = ops(Z,'+',Asum(blk,A,y,iscmpA));    
    ZpATynorm = ops(ZpATy,'norm');
    Rd  = ops(ops(tau,'*',C),'-',ZpATy);
    clear ZpATy;  
    obj = (normb*normC)*[blktrace(blk,C,X) b'*y]/tau; 
    rel_gap = gap/max(1,mean(abs(obj)));
    pri_infeas  = norm(rp(1:m)/tau)/normb2; 
    dual_infeas = ops(ops(Rd,'/',tau),'norm')/normC2; 
  %%pri_infeas  = normb*norm(normA.*rp(1:m)/tau)/normb0;   %% for original data
  %%dual_infeas = normC*ops(ops(Rd,'/',tau),'norm')/normC0;%% 
    infeas_meas = max(pri_infeas,dual_infeas); 
    min_pri_infeas = min(infeashist(1:iter)); 
    gaphist(iter+1)     = gap; 
    infeashist(iter+1)  = infeas_meas; 
    stephist(iter+1) = min(pstep,dstep);
    CPUtime(iter) = cputime-tstart; 
    time(7) = cputime;
    if predcorr; 
       misctime_iter = time(7)-time(6); 
       misctime = misctime + misctime_iter;
    else;
       misctime_iter = time(7)-time(4); 
       misctime = misctime + misctime_iter;
    end; 
%%
    if printyes;
       fprintf('%2.0f  %4.3f %4.3f',iter,pstep,dstep);
       fprintf(' %2.1e %2.1e %2.1e',pri_infeas,dual_infeas,gap);
       fprintf(' %- 7.6e  %3.3f\n',mean(obj),sigma);
    end;
%%
%% check convergence.
%%
     if (mu < 3*eps);
        fprintf('\n Stop: mu is close to machine epsilon');
        termcode = 0;
        break;
     end;
     if (max([rel_gap min_pri_infeas dual_infeas]) < gaptol);
        fprintf('\n Stop: max(relative gap, infeasibilities) < %3.2e',gaptol);
        termcode = 0;
        break;
     end;
     pri_infeas_bad = pri_infeas_bad + ...
        (pri_infeas > max(1e-10,min_pri_infeas) & (min_pri_infeas < 1e-2));
     if (infeas_meas < 1e-4 | pri_infeas_bad) & (mu < 1e-4);      
        idx = [max(1,iter-3): iter]; 
        infeas_slow =  all(infeashist(idx+1)./(1e-10+infeashist(idx)) > 0.7);
        if (mu < 1e-8) | (infeas_slow);
           idx = [max(1,iter-1): iter]; 
        else           
           idx = [max(1,iter-2): iter];
        end;
        gap_ratio  = gaphist(idx+1)./gaphist(idx); 
        idx2 = [max(1,iter-3): iter]; 
        gap_slowrate = max(0.5, 1.2*mean(gaphist(idx2+1)./gaphist(idx2)));
        gap_slow = all(gap_ratio > gap_slowrate) ...
                   & (mu < 1e-3 | rel_gap < 1e-3); 
        if (max(infeashist) > 1e-6) & (mu < 1e-6);
           step_short = all(stephist([iter:iter+1]) < 0.1) | ...
                        (stephist(iter+1) < 0.05); 
        else;
           %% if starting point is nearly feasible, do not test for 
           %% short step here. 
           step_short = 0; 
        end;
        if (gap_slow) | (step_short) | (infeas_slow);  
           fprintf('\n Stop: progress is too slow. ');
           termcode = -5; 
           break;
        end;  
        if (rel_gap < 0.2*infeas_meas) & (mu < 1e-6) 
           fprintf('\n Stop: relative gap < 0.2*infeasibility');
           termcode = 0;
           break; 
        end;
     end;
     if (b'*y > ZpATynorm / max(inftol,1e-13));
        fprintf('\n Stop: primal problem is suspected of being infeasible');
        termcode = 1;
        break;
     end;
     if (-blktrace(blk,C,X) > norm(AX) / max(inftol,1e-13));
        fprintf('\n Stop: dual problem is suspected of being infeasible');
        termcode = 2;
        break;
     end;
     if (mu/mu0 <max(0.1*inftol,1e-13)) & ...
        (tau*kap0/(kap*tau0) <max(0.1*inftol,1e-13));
        infeas_p = (b'*y) / ZpATynorm; 
        infeas_d = (-blktrace(blk,C,X)) / norm(AX); 
        if (infeas_p > infeas_d);
           fprintf('\n Stop: primal problem is possibly infeasible');
           termcode = 1;
        else;
           fprintf('\n Stop: dual problem is possibly infeasible');
           termcode = 2;
        end;
        break;
     end;
  end;
%%
%%---------------------------------------------------------------
%% end of main loop
%%---------------------------------------------------------------
%%
   if (termcode == -6);
      fprintf('\n Stop: maximum number of iterations reached \n');
   end;
%%
%%---------------------------------------------------------------
%% unscale and produce infeasibility certificates if appropriate
%%---------------------------------------------------------------
%%
  if (termcode <= 0),
     if (nargout >= 8); Xiter = X; yiter = y; Ziter = Z; end; 
     if (scale_data == 1),
        X = ops(normb/tau,'*',X); y = (normC/tau)*y./normA;
        Z = ops(normC/tau,'*',Z);
        if (nargout >= 8);
           Xiter = ops(normb,'*',Xiter); yiter = (normC)*yiter./normA;
           Ziter = ops(normC,'*',Ziter);
        end; 
     else;
        X = ops(X,'*',1/tau); y = (1/tau)*y;
        Z = ops(Z,'*',1/tau);
     end;
  end;
  if (termcode == 1);
     if (nargout >= 8); Xiter = X; yiter = y; Ziter = Z; end;
     if (scale_data == 1),
        b = normb*(b.*normA);  y = y./normA;
        X = ops(normb/tau,'*',X);
        if (nargout >= 8);
           Xiter = ops(normb,'*',Xiter); yiter = normC*y; 
           Ziter = ops(normC,'*',Z);
        end; 
     else;
        X = ops(X,'*',1/tau);
     end;
     rby = 1/(b'*y); y = rby*y; Z = ops(Z,'*',rby);
     norm_resid = ZpATynorm * rby;
  end;
  if (termcode == 2);
     if (nargout >= 8); Xiter = X; yiter = y; Ziter = Z; end; 
     if (scale_data == 1),
        C = ops(C,'*',normC);
        y = (normC/tau)*y./normA; Z = ops(normC/tau,'*',Z);
        if (nargout >= 8);
           Xiter = ops(normb,'*',Xiter); yiter = normC*yiter./normA; 
           Ziter = ops(normC,'*',Ziter);
        end;
     else;
        y = (1/tau)*y; Z = ops(Z,'*',1/tau);
     end;
     tCX = blktrace(blk,C,X);
     X = ops(X,'*',1/(-tCX));
     norm_resid = norm(normA.*AX)/(-tCX);
  end;
%%
%%-----------------------------------------
%%  print summary
%%-----------------------------------------
%%
   fprintf('\n------------------------------------\n');
   fprintf(' number of iterations   = %2.0f\n',iter);
   if (termcode <= 0),
      info = [termcode iter gap infeas_meas sum(CPUtime) ];
      fprintf(' gap                    = %3.2e\n',gap);
      fprintf(' relative gap           = %3.2e\n',rel_gap);
      fprintf(' infeasibilities        = %3.2e\n',infeas_meas);
   elseif (termcode == 1),
      info = [termcode iter gap norm_resid sum(CPUtime) ];
      fprintf(' residual of primal infeasibility      \n')
      fprintf(' certificate (y,Z)      = %3.2e\n',norm_resid);
   elseif (termcode == 2),
      info = [termcode iter gap norm_resid sum(CPUtime) ];
      fprintf(' residual of dual infeasibility        \n')
      fprintf(' certificate X          = %3.2e\n',norm_resid);
   end;
   fprintf(' Total CPU time         = %3.1f\n',sum(CPUtime));
   fprintf(' CPU time per iteration = %3.1f\n',sum(CPUtime)/iter);
   fprintf(' termination code       = %2.0f\n',termcode);
   fprintf('----------------------------------------------------\n');
   fprintf('  Percentage of CPU time spent in various parts \n'); 
   fprintf('----------------------------------------------------\n');
   fprintf('  chol   pred   steplen   corr   steplen   misc\n');
   tt = [choltime predtime step_predtime corrtime step_corrtime misctime];
   tt = tt/sum(tt)*100; 
   fprintf('  %3.1f    %3.1f   %3.1f',tt(1),tt(2),tt(3));
   fprintf('       %3.1f   %3.1f       %3.1f\n',tt(4),tt(5),tt(6)); 
   fprintf('----------------------------------------------------\n');
%%=============================================================================
