%%***************************************************************************
%% steplength: compute alpha such that  X + xstep*dX >= 0.
%%
%%  Let X = R'*R, then this is equivalent to 
%%      xstep (R'\dX)/R >= - I  
%%  ie     (R'\dX)/R    >= -(1/xstep) I 
%%  ie   -eig(R'\dX)/R) <= 1/xstep      
%%            
%%  [xstep,d] = steplength(blk,invXchol,dX,randnstate);
%%
%%  Therefore set 1/xstep = max of the left-hand side, and take reciprocal. 
%%  We don't care about negative values (ie positive eigenvalues of R'\dX)/R).
%%  If left hand side <0, step is AWAY from boundary, so set xstep = infinity.
%%                                      
%% SDPT3: version 2.1 
%% Copyright (c) 1997 by
%% K.C. Toh, M.J. Todd, R.H. Tutuncu
%% Last modified: 7 Jul 99
%%***************************************************************************

    function [xstep,d] = steplength(blk,invXchol,dX,randnstate);

%%
    if (nargin == 3); randnstate = 0; end; 
    randn('state',randnstate); 
    use_lanczos = 0; 
%%
    for p = 1:size(blk,1); 
       blktmp = blk{p,2}; 
       if (length(blktmp) == 1); 
          if isreal(dX{p}) & (blktmp >= 200);  
             use_lanczos = 1; 
          elseif ~isreal(dX{p}) & (blktmp >= 50); 
             use_lanczos = 1; 
          end;
       end; 
       if strcmp(blk{p,1},'nondiag') & (use_lanczos); 
          tol = 1e-3;     
          maxit = max(min(blktmp,25),round(sqrt(blktmp))); 
          initvec = randn(sum(blktmp),1);
          [lam,delta,res] = lanczos(invXchol{p},dX{p},maxit,tol,initvec);          
        %%
        %% Note: lam <= actual largest eigenvalue <= lam + delta.
        %%   
          if (res < tol)
             d{p} = lam + delta;
          else
	     d{p} = 1.05*(lam + delta); 
          end 
       else;
          M = -Prod3(blk(p,:),ops(invXchol{p},'ctranspose'),dX{p},...
                              invXchol{p},5); 
          d{p} = blkeig(blk(p,:),M); 
       end;
    end;
    d = ops(d,'real');
%%
    for p = 1:size(blk,1);   
       tmp = max(d{p}) + eps*5*max(abs(d{p})); 
       if (tmp > 0);           
          xstep(p) = 1/max(tmp);
       else
          xstep(p) = 1e6; 
       end;
    end;
    xstep = min(xstep); 
%%===========================================================================
%% lanczos: find the largest eigenvalue of 
%%          invXchol'*dX*invXchol via the lanczos iteration.
%%
%% [lam,delta] = lanczos(invXchol,dX,maxit,tol,v)
%%
%% lam:  an estimate of the largest eigenvalue.
%% lam2: an estimate of the second largest eigenvalue.
%% res:  residual norm of the largest eigen-pair.
%% res2: residual norm of the second largest eigen-pair.
%%
%%***************************************************************************

   function [lam,delta,res] = lanczos(invXchol,dX,maxit,tol,v) 
                                            
   n = length(dX); 
   V = zeros(n,maxit+1); H = zeros(maxit+1,maxit); 
   v = v/norm(v); 
   V(:,1) = v; 
%%
%% lanczos iteration. 
%%
   for k = 1:maxit
      w = dX*(invXchol*v); 
      w = -(w'*invXchol)'; 
      wold = w;
      if (k > 1); 
         w = w - H(k,k-1)*V(:,k-1); 
      end;
      alp = w'*V(:,k); 
      w   = w - alp*V(:,k); 
      H(k,k) = alp; 
      %%
      %% one step of iterative refinement if necessary. 
      %%
      if (norm(w) <= 0.8*norm(wold));
         s = (w'*V(:,1:k))'; 
         w = w - V(:,1:k)*s;
         H(1:k,k) = H(1:k,k) + s;
      end; 
      nrm = norm(w); 
      v = w/nrm; 
      V(:,k+1) = v; 
      H(k+1,k) = nrm;  H(k,k+1) = nrm; 
      %%
      %% compute ritz pairs and test for convergence
      %%
      if (rem(k,5) == 0) | (k == maxit); 
         Hk = H(1:k,1:k); Hk = 0.5*(Hk+Hk'); 
         [Y,D] = eig(Hk); 
         eigH  = real(diag(D)); 
         [dummy,idx] = sort(eigH);
         res_est = abs(H(k+1,k)*Y(k,idx(k)));
         if (res_est <= 0.1*tol) | (k == maxit);
            lam = eigH(idx(k));  
            z   = V(:,1:k)*Y(:,idx(k));
            res = norm(-invXchol'*(dX*(invXchol*z)) - lam*z); 
            lam2 = eigH(idx(k-1)); 
            z2   = V(:,1:k)*Y(:,idx(k-1));
            res2 = norm(-invXchol'*(dX*(invXchol*z2)) - lam*z2);   
            tmp = lam-lam2 - res2; 
            if (tmp > 0); beta = tmp; else; beta = eps; end;  
            delta = min(res,res^2/beta); 
            if (delta <= tol); 
               break; 
            end;
         end; 
      end; 
   end;
%%
%%***************************************************************************
